from __future__ import annotations
from dataclasses import dataclass
import typing
import functools

from xoa_driver.internals.core.builders import (
    build_get_request,
    build_set_request
)
from xoa_driver.internals.core import interfaces
from xoa_driver.internals.core.token import Token
from xoa_driver.internals.core.transporter.registry import register_command
from xoa_driver.internals.core.transporter.protocol.payload import (
    field,
    RequestBodyStruct,
    ResponseBodyStruct,
    XmpByte,
    XmpHex,
    XmpShort,
    XmpInt,
    XmpLong,
    XmpSequence,
    XmpStr,
    XmpMacAddress,
    XmpIPv4Address,
    Hex,
)
from .enums import (
    OnOff,
    IsPermanent,
    AutoOrManual,
    L47TrafficState,
    L47PortState,
    L47PortSpeed,
    IsPresent,
    LicenseSpeed,
    DhcpState,
    DhcpVlanState
)
from . import subtypes


@register_command
@dataclass
class P4_TRAFFIC:
    """
    Gives a traffic state command to a L47 port.
    """

    code: typing.ClassVar[int] = 700
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class SetDataAttr(RequestBodyStruct):
        traffic_state: L47TrafficState = field(XmpByte())
        """coded byte, the traffic state command issued to the port."""

    def set(self, traffic_state: L47TrafficState) -> Token[None]:
        """Set L47 port traffic state.

        :param traffic_state: the traffic state command issued to the port
        :type traffic_state: L47TrafficState
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, traffic_state=traffic_state))

    set_off = functools.partialmethod(set, L47TrafficState.OFF)
    """Set L47 port traffic state to Off."""

    set_on = functools.partialmethod(set, L47TrafficState.ON)
    """Set L47 port traffic state to On."""

    set_stop = functools.partialmethod(set, L47TrafficState.STOP)
    """Set L47 port traffic state to Stop."""

    set_prepare = functools.partialmethod(set, L47TrafficState.PREPARE)
    """Set L47 port traffic state to Prepare."""

    set_prerun = functools.partialmethod(set, L47TrafficState.PRERUN)
    """Set L47 port traffic state to Prerun."""


@register_command
@dataclass
class P4_STATE:
    """
    Display the current state of the L47 port.
    """

    code: typing.ClassVar[int] = 701
    pushed: typing.ClassVar[bool] = True

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        state: L47PortState = field(XmpByte())
        """coded byte, specifying the current state for this port."""

    def get(self) -> Token[GetDataAttr]:
        """Get the current state of the L47 port.

        :return: the current state of the L47 port
        :rtype: P4_STATE.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_CAPABILITIES:
    """
    Report the speeds supported by the L47 port.
    """

    code: typing.ClassVar[int] = 702
    pushed: typing.ClassVar[bool] = True

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        auto: int = field(XmpByte())
        """byte, autoneg supported"""
        N100_mbps: int = field(XmpByte())
        """byte, 100M speed supported"""
        N1_gbps: int = field(XmpByte())
        """byte, 1G speed supported"""
        N2_5_gbps: int = field(XmpByte())
        """byte, 2.5G speed supported"""
        N5_gbps: int = field(XmpByte())
        """byte, 5G speed supported"""
        N10_gbps: int = field(XmpByte())
        """byte, 10G speed supported"""
        N25_gbps: int = field(XmpByte())
        """byte, 25G speed supported"""
        N40_gbps: int = field(XmpByte())
        """byte, 40G speed supported"""
        N50_gbps: int = field(XmpByte())
        """byte, 50G speed supported"""
        N100_gbps: int = field(XmpByte())
        """byte, 100G speed supported"""

    def get(self) -> Token[GetDataAttr]:
        """Get the speeds supported by the L47 port.

        :return: the speeds supported by the L47 port
        :rtype: P4_CAPABILITIES.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_STATE_STATUS:
    """
    Returns status of the last port state change. If the port state has changed to
    PREPARE_FAIL, the status contains information about the reason for the fail.
    Currently the status will be "OK"in all other states.
    """

    code: typing.ClassVar[int] = 703
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        status: str = field(XmpStr())
        """string, status for the last port state change"""

    def get(self) -> Token[GetDataAttr]:
        """Get status of the last port state change.

        :return: status of the last port state change
        :rtype: P4_STATE_STATUS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_VLAN_OFFLOAD:
    """
    Specifies if 802.1Q VLAN tag should be inserted and stripped by the Ethernet
    device. If VLAN Offload is switched ON, VLAN tags will not be present in frames
    captured by the L47 Server.
    """

    code: typing.ClassVar[int] = 704
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        offload: OnOff = field(XmpByte())
        """coded byte, specifies if VLAN Offload is switched ON"""

    class SetDataAttr(RequestBodyStruct):
        offload: OnOff = field(XmpByte())
        """coded byte, specifies if VLAN Offload is switched ON"""

    def get(self) -> Token[GetDataAttr]:
        """Get the VLAN offload status.

        :return: VLAN offload status
        :rtype: P4_VLAN_OFFLOAD.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))

    def set(self, offload: OnOff) -> Token[None]:
        """Set the VLAN offload state.

        :param offload: specifies if VLAN Offload is enabled
        :type offload: OnOff
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, offload=offload))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """Disable VLAN offload."""

    set_on = functools.partialmethod(set, OnOff.ON)
    """Enable VLAN offload."""


@register_command
@dataclass
class P4_ARP_CONFIG:
    """
    Configure the value of the ARP request transmission rate, retransmission timeout
    and max. retries.
    """

    code: typing.ClassVar[int] = 705
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        rate: int = field(XmpInt())
        """integer, ARP Request transmission rate (requests / sec) - must be larger than 0"""
        retrans_timeout: int = field(XmpInt())
        """integer, ARP Request retransmission timeout [ms] - must be larger than 0"""
        retries: int = field(XmpByte())
        """byte, maximum ARP Request retransmission retries"""

    class SetDataAttr(RequestBodyStruct):
        rate: int = field(XmpInt())
        """integer, ARP Request transmission rate (requests / sec) - must be larger than 0"""
        retrans_timeout: int = field(XmpInt())
        """integer, ARP Request retransmission timeout [ms] - must be larger than 0"""
        retries: int = field(XmpByte())
        """byte, maximum ARP Request retransmission retries"""

    def get(self) -> Token[GetDataAttr]:
        """Get the ARP configuration on the port.

        :return: the ARP configuration on the port
        :rtype: P4_ARP_CONFIG.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))

    def set(self, rate: int, retrans_timeout: int, retries: int) -> Token[None]:
        """Set the ARP configuration on the port.

        :param rate: ARP Request transmission rate (requests/sec) - must be larger than 0
        :type rate: int
        :param retrans_timeout: ARP Request retransmission timeout [ms] - must be larger than 0
        :type retrans_timeout: int
        :param retries: maximum ARP Request retransmission retries
        :type retries: int
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, rate=rate, retrans_timeout=retrans_timeout, retries=retries))


@register_command
@dataclass
class P4_NDP_CONFIG:
    """
    Configure the value of the NDP Neighbor Solicitation transmission rate,
    retransmission timeout and max. retries.
    """

    code: typing.ClassVar[int] = 706
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        rate: int = field(XmpInt())
        """integer, NDP Neighbor Solicitation transmission rate (requests / sec) - must be larger than 0"""
        retrans_timeout: int = field(XmpInt())
        """integer, NDP Neighbor Solicitation retransmission timeout [ms] - must be larger than 0"""
        retries: int = field(XmpByte())
        """byte, Max. NDP Neighbor Solicitation retransmission retries"""

    class SetDataAttr(RequestBodyStruct):
        rate: int = field(XmpInt())
        """integer, NDP Neighbor Solicitation transmission rate (requests / sec) - must be larger than 0"""
        retrans_timeout: int = field(XmpInt())
        """integer, NDP Neighbor Solicitation retransmission timeout [ms] - must be larger than 0"""
        retries: int = field(XmpByte())
        """byte, Max. NDP Neighbor Solicitation retransmission retries"""

    def get(self) -> Token[GetDataAttr]:
        """Get the NDP configuration on the port.

        :return: the NDP configuration on the port
        :rtype: P4_NDP_CONFIG.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))

    def set(self, rate: int, retrans_timeout: int, retries: int) -> Token[None]:
        """Set the NDP configuration on the port.

        :param rate: NDP Neighbor Solicitation transmission rate (requests/sec) - must be larger than 0
        :type rate: int
        :param retrans_timeout: NDP Neighbor Solicitation retransmission timeout [ms] - must be larger than 0
        :type retrans_timeout: int
        :param retries: maximum NDP Neighbor Solicitation retransmission retries
        :type retries: int
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, rate=rate, retrans_timeout=retrans_timeout, retries=retries))


@register_command
@dataclass
class P4_CAPTURE:
    """
    Starts or stops packet capture on this port.
    """

    code: typing.ClassVar[int] = 707
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        on_off: OnOff = field(XmpByte())
        """coded byte, specifying whether to capture traffic on this port"""

    class SetDataAttr(RequestBodyStruct):
        on_off: OnOff = field(XmpByte())
        """coded byte, specifying whether to capture traffic on this port"""

    def get(self) -> Token[GetDataAttr]:
        """Get packet capture state on this port.

        :return: packet capture state on this port
        :rtype: P4_CAPTURE.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))

    def set(self, on_off: OnOff) -> Token[None]:
        """Set packet capture state on this port.

        :param on_off: specifying whether to capture traffic on this port
        :type on_off: OnOff
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, on_off=on_off))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """Stop packet capture on this port."""

    set_on = functools.partialmethod(set, OnOff.ON)
    """Start packet capture on this port."""


@register_command
@dataclass
class P4_CAPTURE_GET_FIRST:
    """
    Returns the first captured frame on the port. Command is only valid when port is
    in state STOPPED
    """

    code: typing.ClassVar[int] = 708
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        index: int = field(XmpInt())
        """integer, index of frame returned"""
        second: int = field(XmpInt())
        """integer, second value of frame capture timestamp"""
        microsecond: int = field(XmpInt())
        """integer, microsecond value of frame capture timestamp"""
        capture_length: int = field(XmpInt())
        """integer, length of captured portion of the frame"""
        frame_length: int = field(XmpInt())
        """integer, length of the frame"""
        frame: Hex = field(XmpHex())
        """list of hex bytes, the captured frame (capture_len bytes)"""

    def get(self) -> Token[GetDataAttr]:
        """Get the first captured frame on the port

        :return: the first captured frame on the port
        :rtype: P4_CAPTURE_GET_FIRST.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_CAPTURE_GET_NEXT:
    """
    Returns the next captured frame on the port. Command is only valid when port is
    in state STOPPED
    """

    code: typing.ClassVar[int] = 709
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        index: int = field(XmpInt())
        """integer, index of frame returned"""
        second: int = field(XmpInt())
        """integer, second value of frame capture timestamp"""
        microsecond: int = field(XmpInt())
        """integer, microsecond value of frame capture timestamp"""
        capture_length: int = field(XmpInt())
        """integer, length of captured portion of the frame"""
        frame_length: int = field(XmpInt())
        """integer, length of the frame"""
        frame: Hex = field(XmpHex())
        """hex data, the captured frame (capture_len bytes)"""

    def get(self) -> Token[GetDataAttr]:
        """Get the next captured frame on the port

        :return: the next captured frame on the port
        :rtype: P4_CAPTURE_GET_NEXT.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_ETH_TX_COUNTERS:
    """
    Return total port Ethernet transmit statistics since last clear.
    """

    code: typing.ClassVar[int] = 710
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        current_time: int = field(XmpLong())
        """long integer, the current time (mSec since module restart)"""
        ref_time: int = field(XmpLong())
        """long integer, reference time (mSec for P4_TRAFFIC on)"""
        bits_per_sec: int = field(XmpLong())
        """long integer, bit/second of (layer 2) bytes transmitted"""
        packets_per_sec: int = field(XmpLong())
        """long integer, packets/second of packets transmitted"""
        byte_count: int = field(XmpLong())
        """long integer, total number of (layer 2) bytes transmitted"""
        packet_count: int = field(XmpLong())
        """long integer, total number of packets transmitted"""

    def get(self) -> Token[GetDataAttr]:
        """Get total port Ethernet transmit statistics since last clear.

        :return: total port Ethernet transmit statistics since last clear.
        :rtype: P4_ETH_TX_COUNTERS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_ETH_RX_COUNTERS:
    """
    Return total port Ethernet receive statistics since last clear.
    """

    code: typing.ClassVar[int] = 711
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        current_time: int = field(XmpLong())
        """long integer, the current time (mSec since module restart)"""
        ref_time: int = field(XmpLong())
        """long integer, reference time (mSec for P4_TRAFFIC on)"""
        bits_per_sec: int = field(XmpLong())
        """long integer, bit/second of (layer 2) bytes received"""
        packets_per_sec: int = field(XmpLong())
        """long, integer packets/second of received packets"""
        byte_count: int = field(XmpLong())
        """long integer, total number of (layer 2) bytes received"""
        packet_count: int = field(XmpLong())
        """long integer, total number of packets received"""

    def get(self) -> Token[GetDataAttr]:
        """Get total port Ethernet receive statistics since last clear.

        :return: total port Ethernet receive statistics since last clear.
        :rtype: P4_ETH_RX_COUNTERS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_PORT_TX_COUNTERS:
    """
    Return total port transmit statistics since last clear.
    """

    code: typing.ClassVar[int] = 712
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        current_time: int = field(XmpLong())
        """long integer, the current time (mSec since module restart)"""
        ref_time: int = field(XmpLong())
        """long integer, reference time (mSec for P4_TRAFFIC on)"""
        vlan_packet_count: int = field(XmpLong())
        """long integer, total number of 802.1Q VLAN tagged packets transmitted"""
        bits_per_sec: int = field(XmpLong())
        """long integer, bit/second of (layer 1) bits transmitted."""
        byte_count: int = field(XmpLong())
        """long integer, total number of (layer 1) bytes received."""

    def get(self) -> Token[GetDataAttr]:
        """Get total port transmit statistics since last clear.

        :return: total port transmit statistics since last clear.
        :rtype: P4_PORT_TX_COUNTERS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_PORT_RX_COUNTERS:
    """
    Return total port receive statistics since last clear.
    """

    code: typing.ClassVar[int] = 713
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        current_time: int = field(XmpLong())
        """long integer, the current time (mSec since module restart)"""
        ref_time: int = field(XmpLong())
        """long integer, reference time (mSec for P4_TRAFFIC on)"""
        vlan_packet_count: int = field(XmpLong())
        """long integer, total number of 802.1Q VLAN tagged packets received"""
        bits_per_sec: int = field(XmpLong())
        """long integer, bit/second of (layer 1) bits received."""
        byte_count: int = field(XmpLong())
        """long integer, total number of (layer 1) bytes received."""

    def get(self) -> Token[GetDataAttr]:
        """Get total port receive statistics since last clear.

        :return: total port receive statistics since last clear.
        :rtype: P4_PORT_RX_COUNTERS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_PORT_COUNTERS:
    """
    Return total port transmit error statistics since last clear.
    """

    code: typing.ClassVar[int] = 714
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        current_time: int = field(XmpLong())
        """long integer, the current time (mSec since module restart)"""
        ref_time: int = field(XmpLong())
        """long integer, reference time (mSec for P4_TRAFFIC on)"""
        invalid_eth_count: int = field(XmpLong())
        """long integer, total number of invalid (e.g. short) Ethernet packets received"""
        unknown_eth_count: int = field(XmpLong())
        """long integer, total number of unknown or unsupported Ethernet packets received"""
        mismatch_vlan_error_count: int = field(XmpLong())
        """long integer, total number of packets with mismatching vlan info received"""
        pkt_rate_limit_count: int = field(XmpLong())
        """long integer, number of times that number of packets transmitted has been limited by the maximum packet rate limiter."""

    def get(self) -> Token[GetDataAttr]:
        """Get total port transmit error statistics since last clear.

        :return: total port transmit error statistics since last clear.
        :rtype: P4_PORT_COUNTERS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_TX_PACKET_SIZE:
    """
    Return histogram over transmitted (layer 2) packets sizes in 100 bytes intervals.
    """

    code: typing.ClassVar[int] = 715
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        current_time: int = field(XmpLong())
        """long integer, the current time (mSec since module restart)"""
        ref_time: int = field(XmpLong())
        """long integer, reference time (mSec for P4_TRAFFIC on)"""
        bin_00: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_01: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_02: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_03: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_04: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_05: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_06: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_07: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_08: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_09: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_10: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_11: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_12: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_13: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_14: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_15: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""

    def get(self) -> Token[GetDataAttr]:
        """Get a histogram over transmitted (layer 2) packets sizes in 100 bytes intervals.

        :return: histogram over transmitted (layer 2) packets sizes in 100 bytes intervals.
        :rtype: P4_TX_PACKET_SIZE.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_RX_PACKET_SIZE:
    """
    Return a histogram over received (layer 2) packets sizes in 100 bytes intervals.
    """

    code: typing.ClassVar[int] = 716
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        current_time: int = field(XmpLong())
        """long integer, the current time (mSec since module restart)"""
        ref_time: int = field(XmpLong())
        """long integer, reference time (mSec for P4_TRAFFIC on)"""
        bin_00: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_01: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_02: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_03: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_04: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_05: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_06: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_07: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_08: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_09: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_10: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_11: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_12: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_13: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_14: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""
        bin_15: int = field(XmpLong())
        """long integer, number of packets received with a (layer 2) size in the given interval."""

    def get(self) -> Token[GetDataAttr]:
        """Get a histogram over received (layer 2) packets sizes in 100 bytes intervals.

        :return: a histogram over received (layer 2) packets sizes in 100 bytes intervals.
        :rtype: P4_RX_PACKET_SIZE.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_TX_MTU:
    """
    Return histogram over transmitted (layer 3) packets sizes in 1 byte intervals.
    Each bin represents a packet size in the interval [576..1500] bytes.
    """

    code: typing.ClassVar[int] = 717
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        bins: typing.List[int] = field(XmpSequence(types_chunk=[XmpByte()]))
        """925 x byte, '1' if any packets were transmitted with the specified layer 3 size, otherwise '0'."""

    def get(self) -> Token[GetDataAttr]:
        """Get histogram over transmitted (layer 3) packets sizes in 1 byte intervals.

        :return: histogram over transmitted (layer 3) packets sizes in 1 byte intervals.
        :rtype: P4_TX_MTU.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_RX_MTU:
    """
    Return histogram over received (layer 3) packets sizes in 1 byte intervals. Each
    bin represents a packet size in the interval [576..1500] bytes.
    """

    code: typing.ClassVar[int] = 718
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        bins: typing.List[int] = field(XmpSequence(types_chunk=[XmpByte()]))
        """925 x byte, '1' if any packets were received with the specified layer 3 size, otherwise '0'."""

    def get(self) -> Token[GetDataAttr]:
        """Get histogram over received (layer 3) packets sizes in 1 byte intervals.

        :return: histogram over received (layer 3) packets sizes in 1 byte intervals.
        :rtype: P4_RX_MTU.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_IPV4_RX_COUNTERS:
    """
    Return total Port IPv4 protocol receive statistics since last clear.
    """

    code: typing.ClassVar[int] = 719
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        current_time: int = field(XmpLong())
        """long integer, the current time (mSec since module restart)"""
        ref_time: int = field(XmpLong())
        """long integer, reference time (mSec for P4_TRAFFIC on)"""
        packet_count: int = field(XmpLong())
        """long integer, total number of IPv4 packets received"""

    def get(self) -> Token[GetDataAttr]:
        """Get total Port IPv4 protocol receive statistics since last clear.

        :return: total Port IPv4 protocol receive statistics since last clear.
        :rtype: P4_IPV4_RX_COUNTERS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_IPV4_TX_COUNTERS:
    """
    Return total Port IPv4 protocol transmit statistics since last clear.
    """

    code: typing.ClassVar[int] = 720
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        current_time: int = field(XmpLong())
        """long integer, the current time (mSec since module restart)"""
        ref_time: int = field(XmpLong())
        """long integer, reference time (mSec for P4_TRAFFIC on)"""
        packet_count: int = field(XmpLong())
        """long integer, total number of IPv4 packets transmitted"""

    def get(self) -> Token[GetDataAttr]:
        """Get total Port IPv4 protocol transmit statistics since last clear.

        :return: total Port IPv4 protocol transmit statistics since last clear.
        :rtype: P4_IPV4_TX_COUNTERS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_IPV4_COUNTERS:
    """
    Return total Port IPv4 protocol error statistics since last clear.
    """

    code: typing.ClassVar[int] = 721
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        current_time: int = field(XmpLong())
        """long integer, the current time (mSec since module restart)"""
        ref_time: int = field(XmpLong())
        """long integer, reference time (mSec for P4_TRAFFIC on)"""
        checksum_error_count: int = field(XmpLong())
        """long integer, total number of IPv4 packets which ip header checksum error"""
        invalid_packet_count: int = field(XmpLong())
        """long integer, total number of IPv4 packets which are malformed"""
        unknown_packet_count: int = field(XmpLong())
        """long integer, total number of IPv4 packets with unknown protocol"""

    def get(self) -> Token[GetDataAttr]:
        """Get total Port IPv4 protocol error statistics since last clear.

        :return: total Port IPv4 protocol error statistics since last clear.
        :rtype: P4_IPV4_COUNTERS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_IPV6_RX_COUNTERS:
    """
    Return total Port IPv6 protocol receive statistics since last clear.
    """

    code: typing.ClassVar[int] = 722
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        current_time: int = field(XmpLong())
        """long integer, the current time (mSec since module restart)"""
        ref_time: int = field(XmpLong())
        """long integer, reference time (mSec for P4_TRAFFIC on)"""
        packet_count: int = field(XmpLong())
        """long integer, total number of IPv6 packets received"""

    def get(self) -> Token[GetDataAttr]:
        """Get total Port IPv6 protocol receive statistics since last clear.

        :return: total Port IPv6 protocol receive statistics since last clear.
        :rtype: P4_IPV6_RX_COUNTERS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_IPV6_TX_COUNTERS:
    """
    Return total Port IPv6 protocol transmit statistics since last clear.
    """

    code: typing.ClassVar[int] = 723
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        current_time: int = field(XmpLong())
        """long integer, the current time (mSec since module restart)"""
        ref_time: int = field(XmpLong())
        """long integer, reference time (mSec for P4_TRAFFIC on)"""
        packet_count: int = field(XmpLong())
        """long integer, total number of IPv6 packets transmitted"""

    def get(self) -> Token[GetDataAttr]:
        """Get total Port IPv6 protocol transmit statistics since last clear.

        :return: total Port IPv6 protocol transmit statistics since last clear.
        :rtype: P4_IPV6_TX_COUNTERS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_IPV6_COUNTERS:
    """
    Return total Port IPv6 protocol error statistics since last clear.
    """

    code: typing.ClassVar[int] = 724
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        current_time: int = field(XmpLong())
        """long integer, the current time (mSec since module restart)"""
        ref_time: int = field(XmpLong())
        """long integer, reference time (mSec for P4_TRAFFIC on)"""
        invalid_packet_count: int = field(XmpLong())
        """long integer, total number of ipv6 packets which are malformed"""
        unknown_packet_count: int = field(XmpLong())
        """long integer, total number of ipv6 packets with unknown protocol"""

    def get(self) -> Token[GetDataAttr]:
        """Get total Port IPv6 protocol error statistics since last clear.

        :return: total Port IPv6 protocol error statistics since last clear.
        :rtype: P4_IPV6_COUNTERS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_ARP_RX_COUNTERS:
    """
    Return total Port ARP protocol receive statistics since last clear.
    """

    code: typing.ClassVar[int] = 725
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        current_time: int = field(XmpLong())
        """long integer, the current time (mSec since module restart)"""
        ref_time: int = field(XmpLong())
        """long integer, reference time (mSec for P4_TRAFFIC on)"""
        arp_request_count: int = field(XmpLong())
        """long integer, total number ARP Requests received"""
        arp_reply_count: int = field(XmpLong())
        """long integer, total number ARP Replies received"""

    def get(self) -> Token[GetDataAttr]:
        """Get total Port ARP protocol receive statistics since last clear.

        :return: total Port ARP protocol receive statistics since last clear.
        :rtype: P4_ARP_RX_COUNTERS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_ARP_TX_COUNTERS:
    """
    Return total Port ARP protocol transmit statistics since last clear.
    """

    code: typing.ClassVar[int] = 726
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        current_time: int = field(XmpLong())
        """long integer, the current time (mSec since module restart)"""
        ref_time: int = field(XmpLong())
        """long integer, reference time (mSec for P4_TRAFFIC on)"""
        arp_request_count: int = field(XmpLong())
        """long integer, total number ARP Requests transmitted"""
        arp_reply_count: int = field(XmpLong())
        """long integer, total number ARP Replies transmitted"""

    def get(self) -> Token[GetDataAttr]:
        """Get total Port ARP protocol transmit statistics since last clear.

        :return: total Port ARP protocol transmit statistics since last clear.
        :rtype: P4_ARP_TX_COUNTERS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_ARP_COUNTERS:
    """
    Return total Port ARP protocol error statistics since last clear.
    """

    code: typing.ClassVar[int] = 727
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        current_time: int = field(XmpLong())
        """long integer, the current time (mSec since module restart)"""
        ref_time: int = field(XmpLong())
        """long integer, reference time (mSec for P4_TRAFFIC on)"""
        invalid_arp_count: int = field(XmpLong())
        """long integer, total number of invalid ARP packets received"""
        arp_request_lookup_failure_count: int = field(XmpLong())
        """long integer, number of ARP requests received that could not be resolved"""
        arp_reply_lookup_failure_count: int = field(XmpLong())
        """long integer, number of ARP replies received that could not be resolved"""
        arp_request_retrans_count: int = field(XmpLong())
        """long integer, number of retransmitted ARP requests"""
        arp_resolved_count: int = field(XmpLong())
        """long integer, number of correct resolved IP addresses"""
        arp_failed_count: int = field(XmpLong())
        """long integer, number of IP address that was not resolved"""
        arp_table_lookup_failure_count: int = field(XmpLong())
        """long integer, number of dest IP addresses not found in the ARP table"""

    def get(self) -> Token[GetDataAttr]:
        """Get total Port ARP protocol error statistics since last clear.

        :return: total Port ARP protocol error statistics since last clear.
        :rtype: P4_ARP_COUNTERS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_NDP_RX_COUNTERS:
    """
    Return total Port NDP protocol receive statistics since last clear.
    """

    code: typing.ClassVar[int] = 728
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        current_time: int = field(XmpLong())
        """long integer, the current time (mSec since module restart)"""
        ref_time: int = field(XmpLong())
        """long integer, reference time (mSec for P4_TRAFFIC on)"""
        ndp_request_count: int = field(XmpLong())
        """long integer, total number NDP Requests received"""
        ndp_reply_count: int = field(XmpLong())
        """long integer, total number NDP Replies received"""

    def get(self) -> Token[GetDataAttr]:
        """Get total Port NDP protocol receive statistics since last clear.

        :return: total Port NDP protocol receive statistics since last clear.
        :rtype: P4_NDP_RX_COUNTERS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_NDP_TX_COUNTERS:
    """
    Return total Port NDP protocol transmit statistics since last clear.
    """

    code: typing.ClassVar[int] = 729
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        current_time: int = field(XmpLong())
        """long integer, the current time (mSec since module restart)"""
        ref_time: int = field(XmpLong())
        """long integer, reference time (mSec for P4_TRAFFIC on)"""
        ndp_request_count: int = field(XmpLong())
        """long integer, total number NDP Requests transmitted"""
        ndp_reply_count: int = field(XmpLong())
        """long integer, total number NDP Replies transmitted"""

    def get(self) -> Token[GetDataAttr]:
        """Get total Port NDP protocol transmit statistics since last clear.

        :return: total Port NDP protocol transmit statistics since last clear.
        :rtype: P4_NDP_TX_COUNTERS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_NDP_COUNTERS:
    """
    Return total Port NDP protocol error statistics since last clear.
    """

    code: typing.ClassVar[int] = 730
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        current_time: int = field(XmpLong())
        """long integer, the current time (mSec since module restart)"""
        ref_time: int = field(XmpLong())
        """long integer, reference time (mSec for P4_TRAFFIC on)"""
        invalid_ndp_count: int = field(XmpLong())
        """long integer, total number of invalid NDP packets received"""
        ndp_request_lookup_failure_count: int = field(XmpLong())
        """long integer, number of NDP requests received that could not be resolved"""
        ndp_reply_lookup_failure_count: int = field(XmpLong())
        """long integer, number of NDP replies received that could not be resolved"""
        ndp_request_retrans_count: int = field(XmpLong())
        """long integer, number of retransmitted NDP requests"""
        ndp_resolved_count: int = field(XmpLong())
        """long integer, number of correct resolved IP addresses"""
        ndp_failed_count: int = field(XmpLong())
        """long integer, number of IP address that was not resolved"""
        ndp_table_lookup_failure_count: int = field(XmpLong())
        """long integer, number of dest IP addresses not found in the NDP table"""

    def get(self) -> Token[GetDataAttr]:
        """Get total Port NDP protocol error statistics since last clear.

        :return: total Port NDP protocol error statistics since last clear.
        :rtype: P4_NDP_COUNTERS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_ICMP_RX_COUNTERS:
    """
    Return total Port ICMP protocol receive statistics since last clear.
    """

    code: typing.ClassVar[int] = 731
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        current_time: int = field(XmpLong())
        """long integer, the current time (mSec since module restart)"""
        ref_time: int = field(XmpLong())
        """long integer, reference time (mSec for P4_TRAFFIC on)"""
        icmp_echo_request_count: int = field(XmpLong())
        """long integer, total number of ICMP Echo requests received"""
        icmp_echo_reply_count: int = field(XmpLong())
        """long integer, total number of ICMP Echo replies received"""
        icmp_dest_unknown_count: int = field(XmpLong())
        """long integer, total number of ICMP Destination unknown received"""
        icmp_time_excessive_count: int = field(XmpLong())
        """long integer, total number of ICMP Time exceeded received"""
        icmpv6_count: int = field(XmpLong())
        """long integer, total number of ICMPv6 packets received"""

    def get(self) -> Token[GetDataAttr]:
        """Get total Port ICMP protocol receive statistics since last clear.

        :return: total Port ICMP protocol receive statistics since last clear.
        :rtype: P4_ICMP_RX_COUNTERS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_ICMP_TX_COUNTERS:
    """
    Return total Port ICMP protocol transmit statistics since last clear.
    """

    code: typing.ClassVar[int] = 732
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        current_time: int = field(XmpLong())
        """long integer, the current time (mSec since module restart)"""
        ref_time: int = field(XmpLong())
        """long integer, reference time (mSec for P4_TRAFFIC on)"""
        icmp_echo_request_count: int = field(XmpLong())
        """long integer, total number of ICMP Echo requests transmitted"""
        icmp_echo_reply_count: int = field(XmpLong())
        """long integer, total number of ICMP Echo replies transmitted"""
        icmp_dest_unknown_count: int = field(XmpLong())
        """long integer, total number of ICMP Destination unknown transmitted"""
        icmp_time_excessive_count: int = field(XmpLong())
        """long integer, total number of ICMP Time exceeded transmitted"""
        icmpv6_count: int = field(XmpLong())
        """long integer, total number of ICMPv6 packets transmitted"""

    def get(self) -> Token[GetDataAttr]:
        """Get total Port ICMP protocol transmit statistics since last clear.

        :return: total Port ICMP protocol transmit statistics since last clear.
        :rtype: P4_ICMP_TX_COUNTERS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_ICMP_COUNTERS:
    """
    Return total Port ICMP protocol error statistics since last clear.
    """

    code: typing.ClassVar[int] = 733
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        current_time: int = field(XmpLong())
        """long integer, the current time (mSec since module restart)"""
        ref_time: int = field(XmpLong())
        """long integer, reference time (mSec for P4_TRAFFIC on)"""
        invalid_icmp_count: int = field(XmpLong())
        """long integer, total number of unknown or invalid ICMP packets received"""
        unknown_icmp_count: int = field(XmpLong())
        """long integer, total number of unknown or unsupported ICMP packets received"""
        invalid_icmpv6_count: int = field(XmpLong())
        """long integer, total number of unknown or invalid ICMPv6 packets received"""
        unknown_icmpv6_count: int = field(XmpLong())
        """long integer, total number of unknown or unsupported ICMPv6 packets received"""

    def get(self) -> Token[GetDataAttr]:
        """Get total Port ICMP protocol error statistics since last clear.

        :return: total Port ICMP protocol error statistics since last clear.
        :rtype: P4_ICMP_COUNTERS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_TCP_RX_COUNTERS:
    """
    Return total Port TCP protocol receive statistics since last clear.
    """

    code: typing.ClassVar[int] = 734
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        current_time: int = field(XmpLong())
        """long integer, the current time (mSec since module restart)"""
        ref_time: int = field(XmpLong())
        """long integer, reference time (mSec for P4_TRAFFIC on)"""
        packet_count: int = field(XmpLong())
        """long integer, total number of TCP packets received"""

    def get(self) -> Token[GetDataAttr]:
        """Get total Port TCP protocol receive statistics since last clear.

        :return: total Port TCP protocol receive statistics since last clear.
        :rtype: P4_TCP_RX_COUNTERS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_TCP_TX_COUNTERS:
    """
    Return total Port TCP protocol transmit statistics since last clear.
    """

    code: typing.ClassVar[int] = 735
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        current_time: int = field(XmpLong())
        """long integer, the current time (mSec since module restart)"""
        ref_time: int = field(XmpLong())
        """long integer, reference time (mSec for P4_TRAFFIC on)"""
        packet_count: int = field(XmpLong())
        """long integer, total number of TCP packets transmitted"""

    def get(self) -> Token[GetDataAttr]:
        """Get total Port TCP protocol transmit statistics since last clear.

        :return: total Port TCP protocol transmit statistics since last clear.
        :rtype: P4_TCP_TX_COUNTERS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_TCP_COUNTERS:
    """
    Return total Port TCP protocol error statistics since last clear.
    """

    code: typing.ClassVar[int] = 736
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        current_time: int = field(XmpLong())
        """long integer, the current time (mSec since module restart)"""
        ref_time: int = field(XmpLong())
        """long integer, reference time (mSec for P4_TRAFFIC on)"""
        checksum_error_count: int = field(XmpLong())
        """long integer, total number of tcp packets which tcp header checksum error"""
        invalid_tcp_count: int = field(XmpLong())
        """long integer, total number of TCP packets which are malformed"""
        tcp_lookup_failure_count: int = field(XmpLong())
        """long integer, number of TCP packets received that could not be resolved"""

    def get(self) -> Token[GetDataAttr]:
        """Get total Port TCP protocol error statistics since last clear.

        :return: total Port TCP protocol error statistics since last clear.
        :rtype: P4_TCP_COUNTERS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_UDP_RX_COUNTERS:
    """
    Return total Port UDP protocol receive statistics since last clear.
    """

    code: typing.ClassVar[int] = 737
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        current_time: int = field(XmpLong())
        """long integer, the current time (mSec since module restart)"""
        ref_time: int = field(XmpLong())
        """long integer, reference time (mSec for P4_TRAFFIC on)"""
        packet_count: int = field(XmpLong())
        """long integer, total number of UDP packets received"""

    def get(self) -> Token[GetDataAttr]:
        """Get total Port UDP protocol receive statistics since last clear.

        :return: total Port UDP protocol receive statistics since last clear.
        :rtype: P4_UDP_RX_COUNTERS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_UDP_TX_COUNTERS:
    """
    Return total Port UDP protocol transmit statistics since last clear.
    """

    code: typing.ClassVar[int] = 738
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        current_time: int = field(XmpLong())
        """long integer, the current time (mSec since module restart)"""
        ref_time: int = field(XmpLong())
        """long integer, reference time (mSec for P4_TRAFFIC on)"""
        packet_count: int = field(XmpLong())
        """long integer, total number of UDP packets transmitted"""

    def get(self) -> Token[GetDataAttr]:
        """Get total Port UDP protocol transmit statistics since last clear.

        :return: total Port UDP protocol transmit statistics since last clear.
        :rtype: P4_UDP_TX_COUNTERS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_UDP_COUNTERS:
    """
    Return total Port UDP protocol error statistics since last clear.
    """

    code: typing.ClassVar[int] = 739
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        current_time: int = field(XmpLong())
        """long integer, the current time (mSec since module restart)"""
        ref_time: int = field(XmpLong())
        """long integer, reference time (mSec for P4_TRAFFIC on)"""
        checksum_error_count: int = field(XmpLong())
        """long integer, total number of udp packets which udp header checksum error"""
        invalid_udp_count: int = field(XmpLong())
        """long integer, total number of UDP packets which are malformed"""
        udp_lookup_failure_count: int = field(XmpLong())
        """long integer, number of UDP packets received that could not be resolved"""

    def get(self) -> Token[GetDataAttr]:
        """Get total Port UDP protocol error statistics since last clear.

        :return: total Port UDP protocol error statistics since last clear.
        :rtype: P4_UDP_COUNTERS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_CLEAR_COUNTERS:
    """
    Clears all run-time port counters.
    """

    code: typing.ClassVar[int] = 740
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class SetDataAttr(RequestBodyStruct):
        pass

    def set(self) -> Token[None]:
        """Clears all run-time port counters.
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_ETH_COUNTERS:
    """
    Return total port Ethernet statistics since last clear.
    """

    code: typing.ClassVar[int] = 765
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        current_time: int = field(XmpLong())
        """long integer, the current time (mSec since module restart)"""
        ref_time: int = field(XmpLong())
        """long integer, reference time (mSec for P4_TRAFFIC on)"""
        tx_error_count: int = field(XmpLong())
        """long integer, TX errors"""
        rx_error_count: int = field(XmpLong())
        """long integer, RX errors"""
        rx_packet_lost_count: int = field(XmpLong())
        """long integer, packets lost by the Ethernet driver due to RX queue overflow"""

    def get(self) -> Token[GetDataAttr]:
        """Get total port Ethernet statistics since last clear.

        :return: total port Ethernet statistics since last clear.
        :rtype: P4_ETH_COUNTERS.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_CLEAR:
    """
    Set the Port State to OFF and delete all configured Connection Groups for the port.
    """

    code: typing.ClassVar[int] = 766
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class SetDataAttr(RequestBodyStruct):
        pass

    def set(self) -> Token[None]:
        """Set the Port State to OFF and delete all configured Connection Groups for the port.
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_SPEEDSELECTION:
    """
    Sets the port speed. The selected speed must be one of the speeds supported by
    the port, which can be retrieved with P4_CAPABILITIES.
    """

    code: typing.ClassVar[int] = 767
    pushed: typing.ClassVar[bool] = True

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        speed: L47PortSpeed = field(XmpByte())
        """coded byte, specifies the speed of the port"""

    class SetDataAttr(RequestBodyStruct):
        speed: L47PortSpeed = field(XmpByte())
        """coded byte, specifies the speed of the port"""

    def get(self) -> Token[GetDataAttr]:
        """Get the port speed mode.

        :return: the port speed mode.
        :rtype: P4_SPEEDSELECTION.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))

    def set(self, speed: L47PortSpeed) -> Token[None]:
        """Set the port speed mode.

        :param speed: specifies the speed mode of the port
        :type speed: L47PortSpeed
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, speed=speed))

    set_auto = functools.partialmethod(set, L47PortSpeed.AUTO)
    """Set the port speed mode to Auto."""

    set_f100m = functools.partialmethod(set, L47PortSpeed.F100M)
    """Set the port speed mode to 100 Mbit/s."""

    set_f1g = functools.partialmethod(set, L47PortSpeed.F1G)
    """Set the port speed mode to 1 Gbit/s."""

    set_f2_5g = functools.partialmethod(set, L47PortSpeed.F2_5G)
    """Set the port speed mode to 2.5 Gbit/s."""

    set_f5g = functools.partialmethod(set, L47PortSpeed.F5G)
    """Set the port speed mode to 5 Gbit/s."""

    set_f10g = functools.partialmethod(set, L47PortSpeed.F10G)
    """Set the port speed mode to 10 Gbit/s."""

    set_f25g = functools.partialmethod(set, L47PortSpeed.F25G)
    """Set the port speed mode to 25 Gbit/s."""

    set_f40g = functools.partialmethod(set, L47PortSpeed.F40G)
    """Set the port speed mode to 40 Gbit/s."""

    set_f50g = functools.partialmethod(set, L47PortSpeed.F50G)
    """Set the port speed mode to 50 Gbit/s."""

    set_f100g = functools.partialmethod(set, L47PortSpeed.F100G)
    """Set the port speed mode to 100 Gbit/s."""

@register_command
@dataclass
class P4_ETH_QUEUE_COUNTERS:
    """
    Get the stats of the all active queues of the port
    """

    code: typing.ClassVar[int] = 775
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        last_update:    int = field(XmpLong()) # long integer, the current time (mSec since module restart)
        stat_ref_time:  int = field(XmpLong()) # long integer, reference time (mSec for P4_TRAFFIC on)
        num_queues:     int = field(XmpInt())
        queue_list: typing.List[subtypes.QueueStatsElem] = field(XmpSequence(types_chunk=[XmpLong(), XmpLong(), XmpLong(), XmpLong()]))

    def get(self) -> "Token[GetDataAttr]":
        """Get the stats of the all active queues of the port
        :return: The current time (mSec since module restart)
        :rtype: P4_ETH_QUEUE_COUNTERS.GetDataAttr
        :return: Reference time (mSec for P4_TRAFFIC on)
        :rtype: P4_ETH_QUEUE_COUNTERS.GetDataAttr
        :return: Number of Active queues
        :rtype: P4_ETH_QUEUE_COUNTERS.GetDataAttr
        :return: A list that each contains stats of a queue
        :rtype: P4_ETH_QUEUE_COUNTERS.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))

@register_command
@dataclass
class P4_DHCP_CONFIG:
    """
    Configure DHCP Client in order to aquire a pool of ip addresses.
    """
    
    code: typing.ClassVar[int] = 780
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int

    class SetDataAttr(RequestBodyStruct):
        num_requests:       int = field(XmpInt(signed=False))   # Number of DHCP requests to be sent
        retransmit_retries: int = field(XmpInt(signed=False))   # Max Request retransmission retries
        timeout:            int = field(XmpInt(signed=False))   # retransmission timeout [ms]
        base_hw_addr:       Hex = field(XmpMacAddress())        # base hw address for generating DHCP requests (the first 3 bytes are being used)

    class GetDataAttr(ResponseBodyStruct):
        num_requests:       int = field(XmpInt(signed=False))   # Number of DHCP request to acuire IP addresses
        retransmit_retries: int = field(XmpInt(signed=False))   # Max Request retransmission retries
        timeout:            int = field(XmpInt(signed=False))   # retransmission timeout [ms]
        base_hw_addr:       Hex = field(XmpMacAddress())        # base hw address for generating DHCP requests (the first 3 bytes is being used)

    def get(self) -> "Token[GetDataAttr]":
        """Get the DHCP configuration.

        :return: the DHCP configuration
        :rtype: P4_DHCP_CONFIG.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))

    def set(self, num_requests: int, retransmit_retries: int, timeout: int, base_hw_addr: str) -> "Token":
        """Set the DHCP configuration.

        :param num_requests: Number of DHCP requests - must be larger than 0
        :type num_requests: unsigned int
        :param retransmit_retries: maximum DHCP Request retransmission retries - must be larger than 0
        :type retransmit_retries: unsigned int
        :param timeout: retransmission timeout [ms]
        :type timeout: unsigned int
        :param base_hw_addr: base hw address for generating DHCP requests (the first 3 bytes is being used)
        :type base_hw_addr: str
        """
        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, num_requests=num_requests, 
                                                            retransmit_retries=retransmit_retries, timeout=timeout, base_hw_addr=base_hw_addr))


@register_command
@dataclass
class P4_DHCP_RUN:
    """
    Run DHCP Client process to obtain a pool of address.
    """

    code: typing.ClassVar[int] = 781
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int

    class SetDataAttr(RequestBodyStruct):
        pass

    def set(self) -> "Token":
        """Run DHCP Client Process.

        """
        return Token(self._connection, build_set_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_DHCP_STATE:
    """
    Get the last state of the DHCP Client Process.
    """

    code: typing.ClassVar[int] = 782
    pushed: typing.ClassVar[bool] = False

    _connection: "interfaces.IConnection"
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        dhcp_state:     DhcpState   = field(XmpByte())  # The last state of the DHCP Client Process.
        num_success:    int         = field(XmpInt())   # Current number of successfully optained ip addresses
        num_failure:    int         = field(XmpInt())   # Current number of failed dhcp requests

    def get(self) -> "Token[GetDataAttr]":
        """Get the last state of the DHCP Client Process.
            The dhcp_state is an enum type of DhcpState{DHCP_STATE_UNKNOWN=0, DHCP_STATE_RUNNING=1, DHCP_STATE_COMPLETED=2, DHCP_STATE_FAILED=3}

        :return: the last state of the DHCP Client Process.
        :rtype: P4_DHCP_STATE.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_DHCP_RESULT:
    """
    Get the port's last DHCP client process result
    """

    code: typing.ClassVar[int] = 783
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        dhcp_chunks: typing.List[subtypes.DhcpChunk] = field(XmpSequence(types_chunk=[XmpIPv4Address(), XmpIPv4Address(), XmpIPv4Address(), XmpIPv4Address(), XmpInt(), XmpMacAddress()]))

    def get(self) -> "Token[GetDataAttr]":
        """Get the port's the result of the last DHCP client process

        :return: A list that contains the response for each DHCP request that had lunched by the last DHCP client process 
        :rtype: P4_DHCP_RESULT.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_DHCP_VLAN:
    """
    Configure a set of VLAN tags for current DHCP Process.
    """

    code: typing.ClassVar[int] = 784
    pushed: typing.ClassVar[bool] = True

    _connection: "interfaces.IConnection"
    _module: int
    _port: int
    
    class SetDataAttr(RequestBodyStruct):
        state:  DhcpVlanState = field(XmpByte()) # Enable/Disable Vlan configuration
        vlans:  typing.List[subtypes.VlanTag] = field(XmpSequence(types_chunk=[XmpShort(), XmpByte()])) # A list of vlans. up to 8 TCIs can be define
    
    class GetDataAttr(ResponseBodyStruct):
        state:  DhcpVlanState = field(XmpByte()) # Determines whether VLAN is configured or not.
        vlans:  typing.List[subtypes.VlanTag] = field(XmpSequence(types_chunk=[XmpShort(), XmpByte()])) # A list of vlans. up to 8 TCIs can be define

    def set(self, state: DhcpVlanState, vlans: typing.List[subtypes.VlanTag]) -> "Token":
        """Set a list of VLANs for current DHCP Process, up to 8 TCIs can be defined. The order of the VLANs is like the following: ethernet/vlan0/vlan1/../vlanN/uppler_layer(like IPv4)
            
        :param state: Enable/Disable Vlan configuration
        :type state: DhcpVlanState
        :param vlans: specifying a list of VlanTag
        :type vlans: typing.List[subtypes.VlanTag]
        """
        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, 
                                                            state=state, vlans=vlans))
    def get(self) -> "Token[GetDataAttr]":
        """Get the DHCP VLAN configuration
        :return: Whether DHCP VLAN is configured or not
        :rtype:  P4_DHCP_RESULT.GetDataAttr
        :return: A list of vlans 
        :rtype: P4_DHCP_RESULT.GetDataAttr
        """
        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))

@register_command
@dataclass
class P4_MAX_PACKET_RATE:
    """
    Specifies the maximum number of packets per second allowed to be transmitted on the port.
    """

    code: typing.ClassVar[int] = 950
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        mode: AutoOrManual = field(XmpByte())
        """coded byte, specifies the mode of the max. pps mechanism"""
        rate: int = field(XmpInt())
        """integer, maximum number of packets per second to transmit on this port"""
        time_window: int = field(XmpInt())
        """integer, time window [us] to measure the pps rate"""

    class SetDataAttr(RequestBodyStruct):
        mode: AutoOrManual = field(XmpByte())
        """coded byte, specifies the mode of the max. pps mechanism"""
        rate: int = field(XmpInt())
        """integer, maximum number of packets per second to transmit on this port"""
        time_window: int = field(XmpInt())
        """integer, time window [us] to measure the pps rate"""

    def get(self) -> Token[GetDataAttr]:
        """Get the maximum number of packets per second allowed to be transmitted on the port.

        :return: the maximum number of packets per second allowed to be transmitted on the port.
        :rtype: P4_MAX_PACKET_RATE.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))

    def set(self, mode: AutoOrManual, rate: int, time_window: int) -> Token[None]:
        """Set the maximum number of packets per second allowed to be transmitted on the port.

        :param mode: specifies the mode of the max. pps mechanism
        :type mode: AutoOrManual
        :param rate: maximum number of packets per second to transmit on this port
        :type rate: int
        :param time_window: time window [us] to measure the pps rate
        :type time_window: int
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, mode=mode, rate=rate, time_window=time_window))

    set_automatic = functools.partialmethod(set, AutoOrManual.AUTOMATIC)
    """Set port max packet rate mode to Automatic."""

    set_manual = functools.partialmethod(set, AutoOrManual.MANUAL)
    """Set port max packet rate mode to Manual."""


@register_command
@dataclass
class P4_PCI_INFO:
    """
    Report the port PCI info.
    """

    code: typing.ClassVar[int] = 960
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        vendor_id: Hex = field(XmpHex(size=4))
        """four hex bytes, PCI Vendor ID"""
        device_id: Hex = field(XmpHex(size=4))
        """four hex bytes, PCI Device ID"""
        sub_vendor_id: Hex = field(XmpHex(size=4))
        """four hex bytes, PCI Subsystem Vendor ID"""
        sub_device_id: Hex = field(XmpHex(size=4))
        """four hex bytes, PCI Subsystem Device ID"""
        rev: int = field(XmpInt())
        """integer, Revision"""

    def get(self) -> Token[GetDataAttr]:
        """Get the port PCI info.

        :return: the port PCI info
        :rtype: P4_PCI_INFO.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_FW_VER:
    """
    Report the firmware version of the port (NIC).
    """

    code: typing.ClassVar[int] = 961
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        major: int = field(XmpInt())
        """integer, Major firmware version"""
        minor: int = field(XmpInt())
        """integer, Minor firmware version"""

    def get(self) -> Token[GetDataAttr]:
        """Get the firmware version of the port (NIC).

        :return: the firmware version of the port (NIC)
        :rtype: P4_FW_VER.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_DEV_NAME:
    """
    Report the name of the device (NIC) on which the port is located.
    """

    code: typing.ClassVar[int] = 962
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        name: str = field(XmpStr())
        """string, name of the device (NIC) on which the port is located"""

    def get(self) -> Token[GetDataAttr]:
        """Get the name of the device (NIC) on which the port is located.

        :return: the name of the device (NIC) on which the port is located.
        :rtype: P4_DEV_NAME.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_PORT_TYPE:
    """
    Report the port type. The different possible ports are divided into types.
    """

    code: typing.ClassVar[int] = 963
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        type_number: int = field(XmpInt())
        """integer, enumerated port type"""
        type_string: str = field(XmpStr())
        """string, textual representation of the port type"""

    def get(self) -> Token[GetDataAttr]:
        """Get the L47 port type.

        :return: the L47 port type
        :rtype: P4_PORT_TYPE.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_LICENSE_INFO:
    """
    Returns the information on the license assigned to the port - if any.
    """

    code: typing.ClassVar[int] = 964
    pushed: typing.ClassVar[bool] = True

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        present: IsPresent = field(XmpByte())
        """coded byte, specifies if a license is assigned to the port"""
        speed: LicenseSpeed = field(XmpByte())
        """coded byte, if a license is assigned to the port, specifies the speed of the license"""
        permanency: IsPermanent = field(XmpByte())
        """coded byte, if a license is assigned to the port, specifies if the license is permanent"""
        expiration: int = field(XmpLong())
        """long integer, if a license is assigned to the port and it is not permanent, specifies the expiration date of the license - in seconds since Jan 1, 1970."""

    def get(self) -> Token[GetDataAttr]:
        """Get the information on the license assigned to the port.

        :return: the information on the license assigned to the port
        :rtype: P4_LICENSE_INFO.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))


@register_command
@dataclass
class P4_APTITUDES:
    """
    Returns the ports aptitudes - i.e. what is possible to configure on the port in
    terms of features and performance.

    Current schema of the BSON document:

    .. code-block::

        schema = {
            'chassis': {
                'type': 'int32',
                'required': True,
                'enum': ['CHASSIS_TYPE_UNKNOWN',
                        'CHASSIS_TYPE_APPLIANCE',
                        'CHASSIS_TYPE_BAY',
                        'CHASSIS_TYPE_COMPACT',
                        'CHASSIS_TYPE_SAFIRE']
            },
            'tcp_udp': {
                'type': 'document',
                'required': True,
                'properties': {
                    'cc': {
                        'type': 'int32',
                        'required': True,
                    },
                }
            },
            'tls': {
                'type': 'document',
                'required': True,
                'properties': {
                    'supported': {
                        'type': 'bool',
                        'required': True,
                    },
                    'cc': {
                        'type': 'int32',
                        'required': True,
                    }
                }
            }
        }
    """

    code: typing.ClassVar[int] = 1200
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        bson: typing.List[int] = field(XmpSequence(types_chunk=[XmpByte()]))
        """list of hex bytes, bson document containing the ports aptitudes"""

    def get(self) -> Token[GetDataAttr]:
        """Get the ports aptitudes

        :return: the ports aptitudes in BSON format
        :rtype: P4_APTITUDES.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))



from __future__ import annotations
from dataclasses import dataclass, field
import re
import typing as t
import inspect
import ipaddress
from enum import Enum
from xoa_driver.internals import commands
from xoa_driver import enums
from xoa_driver.testers import GenericAnyTester
from xoa_driver.modules import GenericAnyModule
from xoa_driver.ports import GenericAnyPort
from xoa_driver.misc import ArpChunk, NdpChunk
from xoa_driver.utils import apply_iter
from xoa_driver.internals.core.token import Token
from xoa_driver.internals.core.transporter.protocol.payload import Hex
from xoa_driver.internals.core.transporter.protocol.struct_request import Request
from xoa_driver.internals.core.transporter.protocol._constants import CommandType
from xoa_driver.internals.core.transporter._typings import (XoaCommandType, ICmdOnlyGet, ICmdOnlySet)


class ICmdOnlyGett(ICmdOnlyGet, t.Protocol):
    __name__: t.ClassVar[str]


class ICmdOnlySett(ICmdOnlySet, t.Protocol):
    __name__: t.ClassVar[str]


def build_set_requestt(cls: ICmdOnlySett, **kwargs) -> Request:
    indices = kwargs.pop("indices", [])
    module = kwargs.pop("module", None)
    port = kwargs.pop("port", None)
    req_values = cls.SetDataAttr(**kwargs)
    return Request(class_name=cls.__name__, cmd_type=CommandType.COMMAND_VALUE, cmd_code=cls.code, module_index=module, port_index=port, indices=indices, values=req_values)


def build_get_requestt(cls: ICmdOnlyGett, **kwargs) -> Request:
    indices = kwargs.pop("indices", [])
    module = kwargs.pop("module", None)
    port = kwargs.pop("port", None)
    req_values = None
    return Request(class_name=cls.__name__, cmd_type=CommandType.COMMAND_QUERY, cmd_code=cls.code, module_index=module, port_index=port, indices=indices, values=req_values)


module = port = r"\d+"
index = r"((0x|0X)?[A-Fa-f\d]+)"
zero_or_more_space = r"\s*"
module_port_group = r"((?P<module>\d+)(/(?P<port>\d+))?\s*)?"
command_name_group = r"(?P<command_name>[A-Z_a-z0-9]+\s*)"
indices_group = r"(?P<indices>(\[((0x|0X)?[A-Fa-f\d]+)(,?\s*((0x|0X)?[A-Fa-f\d]+))?(,?\s*((0x|0X)?[A-Fa-f\d]+))?\]\s*)?)"
params_group = r"(?P<params>.*)"
command_pattern = re.compile(module_port_group + command_name_group + indices_group + params_group)


class Unpassed:
    """
    None has special meaning, so this empty type is needed for parameters which are not passed.
    """

    pass


@dataclass
class Body:
    command_name: str = ""
    cmd_class: XoaCommandType = XoaCommandType  # type: ignore
    type: CommandType = CommandType.COMMAND_STATUS
    module: t.Optional[int] = None
    port: t.Optional[int] = None
    square_indices: list[int] = field(default_factory=list)
    values: dict = field(default_factory=dict)

    def as_request(
        self,
        module_num: int | None | Unpassed = Unpassed(),
        port_num: int | None | Unpassed = Unpassed(),
        indices_num: list[int] | Unpassed = Unpassed(),
    ) -> Request:
        if not isinstance(indices_num, Unpassed):
            self.square_indices = indices_num
        if not isinstance(module_num, Unpassed):
            self.module = module_num
        if not isinstance(port_num, Unpassed):
            self.port = port_num

        dic = dict(indices=self.square_indices, module=self.module, port=self.port, **self.values,)
        return build_get_requestt(self.cmd_class, **dic) if self.type == CommandType.COMMAND_QUERY else build_set_requestt(self.cmd_class, **dic)  # type: ignore


class CLIConverter:
    @classmethod
    def _read_indices(cls, indices_str: str) -> list[int]:
        result = []
        ind = indices_str.strip().strip("[").strip("]").split(",")
        for i in ind:
            if not i:
                continue
            try:
                if "0x" in i or "0X" in i:
                    result.append(int(i.strip(), 16))
                else:
                    result.append(int(i.strip()))
            except Exception:
                raise ValueError(f"Invalid indices str {indices_str}!")
        return result

    @classmethod
    def _special_read(cls, class_name: str, or_params: list[str]) -> list[str]:
        params = or_params
        if class_name == "C_DOWN" and params.index("-1480937026") == 0:
            params.pop(0)
        elif class_name == "P_ARPRXTABLE":
            params = []
            if or_params:
                orr_params = or_params[0].replace("0x", "").replace("0X", "")
                params = [orr_params[i: i + 26] for i in range(0, len(orr_params), 26)]
        elif class_name == "P_NDPRXTABLE":
            params = []
            if or_params:
                orr_params = or_params[0].replace("0x", "").replace("0X", "")
                params = [orr_params[i: i + 50] for i in range(0, len(orr_params), 50)]

        return params

    @classmethod
    def _special_add(cls, class_name: str, dic: dict) -> dict:
        if class_name in ("C_DOWN", "C_FILEFINISH", "M_UPGRADE"):
            dic["magic"] = -1480937026
        elif class_name == "M_FPGAREIMAGE":
            dic["key_code"] = 42
        elif class_name == "PP_EYEMEASURE":
            dic["dummy"] = []
        elif class_name == "PP_PHYTXEQ":
            dic["mode"] = 4
        return dic

    @classmethod
    def _read_response_values(cls, or_params: list[str], cmd_class: t.Type) -> dict:
        dic = {}
        attr_set = getattr(cmd_class, "set")
        class_name = cmd_class.__name__
        values = list(inspect.signature(attr_set).parameters.values())
        params = cls._special_read(class_name, or_params)
        func_sigs: list[inspect.Parameter] = []
        list_index = -1
        for i, v in enumerate(values):
            if v.name == "self":
                continue
            if "list[" in str(v.annotation).lower():
                list_index = i - 1
            func_sigs.append(v)

        if list_index == -1:
            dic = {f.name: cls._bind_one_param(class_name, p, f.annotation, f.name) for p, f in zip(params, func_sigs)}
        else:
            element_sig = (values[list_index + 1].annotation.replace("typing.", "").replace("list[", "").replace("List[", "").replace("]", ""))
            dic = cls._bind_has_list(class_name, params, func_sigs, list_index, element_sig)
        dic = cls._special_add(class_name, dic)
        return dic

    @classmethod
    def _bind_has_list(cls, class_name: str, params: list[str], func_sigs: list[inspect.Parameter], list_index: int, element_sig: str,) -> dict[str, t.Any]:
        dic = {}
        name = func_sigs[list_index].name
        fore_sigs = func_sigs[:list_index]
        fore_params = params[:list_index]

        back_sigs = func_sigs[list_index + 1:]
        back_params = params[-len(back_sigs):]

        if len(func_sigs) == 1:  # only one param and is list
            dic.update({name: [cls._bind_one_param(class_name, p, element_sig) for p in params]})
        elif len(fore_params) == len(fore_sigs) != 0:  # others followed by a list
            list_params = params[list_index:]
            dic.update({f.name: cls._bind_one_param(class_name, p, f.annotation, f.name) for p, f in zip(fore_params, fore_sigs)})
            dic[name] = [cls._bind_one_param(class_name, b, element_sig) for b in list_params]
        elif len(back_params) == len(back_sigs) != 0:  # a list follow by others
            list_params = params[: -(len(func_sigs) - list_index - 1)]
            dic.update({f.name: cls._bind_one_param(class_name, p, f.annotation, f.name) for p, f in zip(back_params, back_sigs)})
            dic[name] = [cls._bind_one_param(class_name, b, element_sig) for b in list_params]
        return dic

    @classmethod
    def _basic_cast(
        cls, string_param: str, type_name: str
    ) -> str | int | float | Hex | None:
        basics = {"str": str, "int": int, "float": float, "Hex": Hex}
        basic_type = basics.get(type_name, None)
        if basic_type is not None:
            try:
                if basic_type is str:
                    return str(string_param).strip('"').strip("'")
                elif basic_type is Hex:
                    s = string_param.replace("0x", "").replace("0X", "")
                    if len(s) % 2:
                        s = "0" + s
                    return Hex(s)
                return basic_type(string_param)
            except ValueError:
                return None

    @classmethod
    def _ip_cast(
        cls,
        string_param: str,
        type_name: str,
    ) -> ipaddress.IPv4Address | ipaddress.IPv6Address | None:
        ips = {"ipaddress.IPv4Address": ipaddress.IPv4Address, "IPv4Address": ipaddress.IPv4Address,
               "ipaddress.IPv6Address": ipaddress.IPv6Address, "IPv6Address": ipaddress.IPv6Address}
        ips_cast = ips.get(type_name, None)
        if ips_cast is not None:
            try:
                return ips_cast(string_param)
            except Exception:
                pass
            try:
                # input is int-IPv4 but in a str form:
                # ipaddress.ip_address(3232235788) -> IPv4Address('192.168.1.12')
                return ips_cast(int(string_param))
            except Exception:
                pass
            try:
                # input is IPV6 but in a str form:
                # ipaddress.ip_address(int('0x10000000000000000000000000000000'), 16) -> IPv6Address('1000::')
                return ips_cast(int(string_param, 16))
            except Exception:
                pass
        return None

    @classmethod
    def _enum_cast(cls, string_param: str, type_name: str) -> Enum | None:
        enum_cast = getattr(enums, type_name, None)
        if enum_cast is not None:
            try:
                return enum_cast(int(string_param))
            except Exception:
                pass
            try:
                return enum_cast[string_param.upper()]
            except Exception:
                pass
        return None

    @classmethod
    def _special_cast(cls, class_name: str, string_param: str, type_name: str, param_name: str) -> t.Any:
        if type_name == "ProtocolOption":
            enum_cast = getattr(enums, type_name, None)
            if enum_cast is not None:
                try:
                    i = int(string_param)
                    return enum_cast(256 + i)
                except Exception:
                    return None
        elif type_name == "SMAInputFunction":
            enum_cast = getattr(enums, type_name, None)
            if enum_cast is not None and string_param == "NOTUSED":
                return enum_cast["NOT_USED"]
        elif type_name == "SourceType":
            enum_cast = getattr(enums, type_name, None)
            if enum_cast is not None:
                return {"TXIFG": enum_cast["TX_IFG"], "TXLEN": enum_cast["TX_LEN"],
                        "RXIFG": enum_cast["RX_IFG"], "RXLEN": enum_cast["RX_LEN"],
                        "RXLAT": enum_cast["RX_LATENCY"], "RXJIT": enum_cast["RX_JITTER"]}.get(string_param, None)
        elif type_name == "LinkTrainingMode":
            enum_cast = getattr(enums, type_name, None)
            if enum_cast is not None and string_param == "AUTO":
                return enum_cast["START_AFTER_AUTONEG"]
        elif type_name == "TimeoutMode":
            enum_cast = getattr(enums, type_name, None)
            if enum_cast is not None and string_param == "DEFAULT_TIMEOUT":
                return enum_cast["DEFAULT"]
        elif class_name == "P_MULTICASTHDR":
            enum_cast = getattr(enums, type_name, None)
            if enum_cast is not None:
                return enum_cast[string_param.upper().replace("DEI_", "")]
        elif class_name == "P_CHECKSUM":
            return {"ON": 14, "OFF": 0}.get(string_param, None)
        elif class_name == "P_ARPRXTABLE":
            ipv4_address = ipaddress.IPv4Address(bytes.fromhex(string_param[0:8]))
            prefix = int.from_bytes(bytes.fromhex(string_param[8:12]), "big")
            patched_mac = enums.OnOff(int.from_bytes(bytes.fromhex(string_param[12:14]), "big"))
            mac_address = Hex(string_param[14:])
            chunk = ArpChunk(ipv4_address, prefix, patched_mac, mac_address)
            return chunk
        elif class_name == "P_NDPRXTABLE":
            ipv6_address = ipaddress.IPv6Address(bytes.fromhex(string_param[0:32]))
            prefix = int.from_bytes(bytes.fromhex(string_param[32:36]), "big")
            patched_mac = enums.OnOff(int.from_bytes(bytes.fromhex(string_param[36:38]), "big"))
            mac_address = Hex(string_param[38:])
            chunk = NdpChunk(ipv6_address, prefix, patched_mac, mac_address)
            return chunk
        elif class_name == "PL1_LINKTRAIN_CMD":
            s = string_param.upper()
            t1 = ("PRE1", "MAIN", "POST", "PRE2", "PRE3")
            t2 = ("PRESET_1", "PRESET_2", "PRESET_3", "PRESET_4", "PRESET_5")
            t3 = ("PAM2", "PAM4", "PAM4_WITH_PRECODING")
            if s in t1:
                return t1.index(s)
            elif s in t2:
                return t2.index(s)
            elif s in t3:
                return t3.index(s)
        elif class_name == "PP_LINKFLAP_PARAMS" and param_name == "repetition":
            if string_param.upper() == "INFINITE":
                return 0
            else:
                return int(string_param)
            

    @classmethod
    def _bind_one_param(
        cls, class_name: str, string_param: str, type_name: str, param_name: str = ""
    ) -> t.Any:
        basic_cast = cls._basic_cast(string_param, type_name)
        if basic_cast is not None:
            return basic_cast

        ip_cast = cls._ip_cast(string_param, type_name)
        if ip_cast is not None:
            return ip_cast

        enum_cast = cls._enum_cast(string_param, type_name)
        if enum_cast is not None:
            return enum_cast

        special_cast = cls._special_cast(class_name, string_param, type_name, param_name)
        if special_cast is not None:
            return special_cast

        raise ValueError(f"Cannot bind param '{string_param}' to type '{type_name}'!")

    @classmethod
    def _parse_param_string(cls, param_string: str) -> list[str]:
        """Handle extreme cases where \',\",\\\',\\\" or string inside param_string."""
        (
            OUTSIDE,
            INSIDE_DOUBLE_QUOTE,
            INSIDE_SINGLE_QUOTE,
            ESCAPED_DOUBLE_QUOTE,
            ESCAPED_SINGLE_QUOTE,
        ) = range(5)
        state = OUTSIDE
        string_list = []
        buf = []
        for c in param_string:
            if state == OUTSIDE:
                if c in (" ", "\t", "\r", "\n", "\f"):
                    string = "".join(buf).strip()
                    if string:
                        string_list.append(string)
                    if buf:
                        buf = []
                elif c == '"':
                    state = INSIDE_DOUBLE_QUOTE
                elif c == "'":
                    state = INSIDE_SINGLE_QUOTE
                else:
                    buf.append(c)
            elif state == INSIDE_DOUBLE_QUOTE:
                if c == "\\":
                    state = ESCAPED_DOUBLE_QUOTE
                elif c == '"':
                    string_list.append("".join(buf))
                    buf = []
                    state = OUTSIDE
                else:
                    buf.append(c)
            elif state == INSIDE_SINGLE_QUOTE:
                if c == "\\":
                    state = ESCAPED_SINGLE_QUOTE
                elif c == "'":
                    string_list.append("".join(buf))
                    buf = []
                    state = OUTSIDE
                else:
                    buf.append(c)
            elif state == ESCAPED_DOUBLE_QUOTE:
                buf.append(c)
                state = INSIDE_DOUBLE_QUOTE
            elif state == ESCAPED_SINGLE_QUOTE:
                buf.append(c)
                state = INSIDE_SINGLE_QUOTE
        if state != OUTSIDE:
            s = '"' if state == INSIDE_DOUBLE_QUOTE else "'"
            raise ValueError(f"String not complete: ({s}{''.join(buf)}).")
        if buf:
            string_list.append("".join(buf).strip())
        return string_list

    @classmethod
    def _read_params(
        cls, param_string: str, cmd_class: t.Type
    ) -> tuple[CommandType, dict]:
        params = cls._parse_param_string(param_string)
        if params == ["?"]:
            return CommandType.COMMAND_QUERY, {}
        return CommandType.COMMAND_VALUE, cls._read_response_values(params, cmd_class)

    @classmethod
    def _read_command_name(cls, command_name: str) -> t.Type:
        return getattr(commands, command_name.upper())

    @classmethod
    def _read_int(cls, n: str | int | None) -> int | None:
        if n is None:
            return None
        try:
            if isinstance(n, str) and ("0x" in n or "0X" in n):
                return int(n.strip(), 16)
            return int(n)
        except Exception:
            return None

    @classmethod
    def convert_each_command(cls, command: str) -> Body | None:
        result = re.search(command_pattern, command)
        if not result:
            return None
        command_name = result.group("command_name").strip() or ""
        cmd_class = cls._read_command_name(command_name)
        module_index = cls._read_int(result.group("module"))
        port_index = cls._read_int(result.group("port"))
        indices = cls._read_indices(result.group("indices"))
        body_type, values = cls._read_params(result.group("params") or "", cmd_class)
        body = Body(command_name, cmd_class, body_type, module_index, port_index, indices, values)
        return body

    @classmethod
    def read_commands_from_file(
        cls, path: str, comment_start: tuple[str, ...] = (";", "#", "//")
    ) -> t.Generator[Body, None, None]:
        with open(path, "r") as f:
            for line in f:
                buf = line.strip()
                for i, char in enumerate(buf):
                    if char in comment_start:
                        buf = buf[:i]
                        break
                if not buf:
                    continue
                r = cls.convert_each_command(buf)
                if r is not None:
                    yield r

    @classmethod
    def read_commands_from_string(
        cls, long_str: str, comment_start: tuple[str, ...] = (";", "#", "//")
    ) -> t.Generator[Body, None, None]:
        for line in long_str.split("\n"):
            buf = line.strip()
            for i, char in enumerate(buf):
                if char in comment_start:
                    buf = buf[:i]
                    break
            if not buf:
                continue
            r = cls.convert_each_command(buf)
            if r is not None:
                yield r


read_commands_from_file = CLIConverter.read_commands_from_file
read_commands_from_string = CLIConverter.read_commands_from_string


def upload_config_from(obj: GenericAnyTester | GenericAnyModule | GenericAnyPort, long_str: str,
                       is_file: bool, mode: str, comment_start: tuple[str, ...] = (";", "#", "//")) -> t.Generator[Token, None, None]:
    func = read_commands_from_file if is_file else read_commands_from_string
    for command in func(long_str, comment_start):
        if not command.command_name.startswith(mode):
            continue
        if mode == "M":
            request = command.as_request(module_num=getattr(obj, "module_id", None))
        elif mode == "P":
            request = command.as_request(
                module_num=getattr(getattr(obj, "kind", None), "module_id", None),
                port_num=getattr(getattr(obj, "kind", None), "port_id", None),
            )
        else:
            request = command.as_request()
        yield Token(obj._conn, request)


async def _helper(obj: GenericAnyTester | GenericAnyModule | GenericAnyPort, long_str: str, is_file: bool, mode: str, comment_start: tuple[str, ...]) -> None:
    async for f in apply_iter(*upload_config_from(obj, long_str, is_file, mode, comment_start), return_exceptions=True):
        pass


async def tester_config_from_string(tester: GenericAnyTester, long_str: str, comment_start: tuple[str, ...] = (";", "#", "//")) -> None:
    """Send tester configuration from a string. The CLI commands must all start with `C_` prefix.

    :param tester: the tester object
    :type tester: GenericAnyTester
    :param long_str: the string containing the CLI commands
    :type long_str: str
    :param comment_start: symbol used to start a comment, defaults to (";", "#", "//")
    :type comment_start: tuple[str, ...], optional
    """
    await _helper(tester, long_str, False, "C", comment_start)


async def tester_config_from_file(tester: GenericAnyTester, path: str, comment_start: tuple[str, ...] = (";", "#", "//")) -> None:
    """Send tester configuration from a configuration file. The CLI commands must all start with `C_` prefix.

    :param tester: the tester object
    :type tester: GenericAnyTester
    :param path: the path to the configuration file
    :type path: str
    :param comment_start: symbol used to start a comment, defaults to (";", "#", "//")
    :type comment_start: tuple[str, ...], optional
    """
    await _helper(tester, path, True, "C", comment_start)


async def module_config_from_string(module: GenericAnyModule, long_str: str, comment_start: tuple[str, ...] = (";", "#", "//")) -> None:
    """Send module configuration from a string. The CLI commands must all start with `M_` prefix.

    :param module: the module object
    :type module: GenericAnyModule
    :param long_str: the string containing the CLI commands
    :type long_str: str
    :param comment_start: symbol used to start a comment, defaults to (";", "#", "//")
    :type comment_start: tuple[str, ...], optional
    """
    assert (module.is_reserved_by_me()), f"Please reserve Module {module.module_id} first!"
    await _helper(module, long_str, False, "M", comment_start)


async def module_config_from_file(module: GenericAnyModule, path: str, comment_start: tuple[str, ...] = (";", "#", "//")) -> None:
    """Send module configuration from a configuration file. The CLI commands must all start with `M_` prefix.

    :param module: the module object
    :type module: GenericAnyModule
    :param path: the path to the configuration file
    :type path: str
    :param comment_start: symbol used to start a comment, defaults to (";", "#", "//")
    :type comment_start: tuple[str, ...], optional
    """
    assert (module.is_reserved_by_me()), f"Please reserve Module {module.module_id} first!"
    await _helper(module, path, True, "M", comment_start)


async def port_config_from_string(port: GenericAnyPort, long_str: str, comment_start: tuple[str, ...] = (";", "#", "//")) -> None:
    """Send port configuration from a string. The CLI commands must all start with `P_` prefix.

    :param port: the port object
    :type port: GenericAnyPort
    :param long_str: the string containing the CLI commands
    :type long_str: str
    :param comment_start: symbol used to start a comment, defaults to (";", "#", "//")
    :type comment_start: tuple[str, ...], optional
    """
    assert (port.is_reserved_by_me()), f"Please reserve Port {port.kind.module_id}/{port.kind.port_id} first!"
    await _helper(port, long_str, False, "P", comment_start)


async def port_config_from_file(port: GenericAnyPort, path: str, comment_start: tuple[str, ...] = (";", "#", "//")) -> None:
    """Send port configuration from a port configuration file (.xpc file).

    :param port: the port object
    :type port: GenericAnyPort
    :param path: the path to the configuration file
    :type path: str
    :param comment_start: symbol used to start a comment, defaults to (";", "#", "//")
    :type comment_start: tuple[str, ...], optional
    """
    assert (port.is_reserved_by_me()), f"Please reserve Port {port.kind.module_id}/{port.kind.port_id} first!"
    await _helper(port, path, True, "P", comment_start)


__all__ = (
    "read_commands_from_file",
    "read_commands_from_string",
    "tester_config_from_string",
    "module_config_from_string",
    "port_config_from_string",
    "tester_config_from_file",
    "module_config_from_file",
    "port_config_from_file",
)



from __future__ import annotations
from xoa_driver.internals.commands import (
    C_BUILDSTRING,
    C_VERSIONNO_MINOR
)
from xoa_driver.internals.core.transporter.logger import CustomLogger
from .l47_tester import L47Tester


class L47VeTester(L47Tester):
    """
    This is a conceptual class of Xena VulcanVE Tester.
    It is essentially an extended :class:`BaseTester`.


    :param host: tester's address/hostname
    :type host: str
    :param username: username of the user
    :type username: str
    :param password: login password of the tester, defaults to "xena"
    :type password: str, optional
    :param port: the port number for connection establishment, default to 22606
    :type port: int, optional
    :param enable_logging: `True` if debug log output from the tester is needed, and `False` otherwise
    :type enable_logging: bool
    :param custom_logger: `None` allow to specify logger with custom configurations
    :type enable_logging: CustomLogger | None
    :param session_timeout: `130` Session timeout in seconds
    :type session_timeout: int
    """

    def __init__(self, host: str, username: str, password: str = "xena", port: int = 22606, *, enable_logging: bool = False, custom_logger: CustomLogger | None = None, session_timeout: int = 130) -> None:
        super().__init__(
            host=host,
            username=username,
            password=password,
            port=port,
            enable_logging=enable_logging,
            custom_logger=custom_logger,
            session_timeout=session_timeout
        )
        self.version_no_minor = C_VERSIONNO_MINOR(self._conn)
        """
        Get the minor version number of the tester firmware.

        :type: C_VERSIONNO_MINOR
        """

        self.build_string = C_BUILDSTRING(self._conn)
        """
        Identify the hostname of the PC that builds the xenaserver. It uniquely identifies the build of a xenaserver.

        :type: C_BUILDSTRING
        """



#: High-level functions
from xoa_driver.functions import (
    anlt,
    cli,
    mgmt,
    anlt_ll_debug,
    exceptions,
)

__all__ = (
    "anlt",
    "mgmt",
    "anlt_ll_debug",
    "exceptions",
    "cli"
)


import ipaddress
from dataclasses import dataclass

from xoa_driver.internals.core.transporter.protocol.payload import (
    field,
    XmpByte,
    XmpShort,
    XmpInt,
    XmpLong,
    XmpMacAddress,
    XmpIPv4Address,
    Hex,
)
from .enums import (
    OnOff,
    VlanType
)


@dataclass
class ArpChunk:
    ipv4_address: ipaddress.IPv4Address
    """The IP address to match to the Target IP address in the ARP requests"""

    prefix: int
    """The IP prefix used for address matching."""

    patched_mac: OnOff
    """If on, the target MAC address will be patched with the part of the IP address that is not masked by the prefix"""

    mac_address: Hex
    """The target MAC address to return in the ARP reply. If set t to all-zero, the port MAC address will be used."""


@dataclass
class NdpChunk:
    ipv6_address: ipaddress.IPv6Address
    """The IP address to match to the Target IP address in the NDP Neighbor Solicitation"""

    prefix: int
    """The IP prefix used for address matching."""

    patched_mac: OnOff
    """If on, the target MAC address will be patched with the part of the IP address that is not masked by the prefix"""

    mac_address: Hex
    """The target MAC address to return in the NDP Neighbor Advertisement. If set t to all-zero, the port MAC address will be used."""

@dataclass
class DhcpChunk:
    ipv4_address:       ipaddress.IPv4Address = field(XmpIPv4Address())  # provided IP address by DHCP Server
    subnet_mask:        ipaddress.IPv4Address = field(XmpIPv4Address())  # provided Subnet mask by DHCP Server
    braodcast_address:  ipaddress.IPv4Address = field(XmpIPv4Address())  # provided Broadcast address by DHCP Server
    defaultgw:          ipaddress.IPv4Address = field(XmpIPv4Address())  # provided Default GW by DHCP Server
    lease_time:         int                   = field(XmpInt())          # provided Lease Time by DHCP Server
    mac_address:        Hex                   = field(XmpMacAddress())   # The client source mac address that this DHCP request blongs to

@dataclass
class GroupAddressElem:
    def __init__(self, ip_address, subnet_mask, mac_address):
        self.ipv4_address = ip_address
        self.subnet_mask = subnet_mask
        self.mac_address = mac_address
        
    
    
    ipv4_address:   ipaddress.IPv4Address = field(XmpIPv4Address()) # Ip address
    subnet_mask:    ipaddress.IPv4Address = field(XmpIPv4Address()) # Subne Mask
    mac_address:    Hex                   = field(XmpMacAddress())  # Mac Address

    
@dataclass
class VlanTag:
    def __init__(self, tci, type):
        self.tci = tci
        self.type = type
    
    tci: int        =   field(XmpShort())  # 
    type: VlanType  =   field(XmpByte())  # Subne Mask
    
@dataclass
class QueueStatsElem:
    rx_bytes:   int = field(XmpLong())  # Total Received Bytes
    rx_packets: int = field(XmpLong())  # Total Received Packets
    tx_bytes:   int = field(XmpLong())  # Total Sent Bytes
    tx_packets: int = field(XmpLong())  # Total Sent Packets



from __future__ import annotations
import typing as t
import logging

from .__state_off import StateOff
from .__state_on_default import StateOnDefault
from .__state_on_loguru import StateOnLoguru
from .__state_on_user import StateOnUser


class TransportationLogger:
    __slots__ = ("enabled", "identity_name", "_logger", "__state")

    def __init__(self, cid: str, enabled: bool = False, logger: CustomLogger | None = None) -> None:
        self.enabled = enabled
        self.identity_name = f"{cid}"
        self.__state = self.__select_state(logger)
        self._logger = self.__state.setup(self.identity_name, logger)

    def __select_state(self, logger) -> t.Type[LoggerState]:
        if not self.enabled:
            return StateOff
        elif self.enabled and logger is None:
            return StateOnDefault
        elif isinstance(logger, logging.Logger):
            return StateOnDefault
        try:
            from loguru import Logger as LoguruLogger
        except ImportError:
            pass
        else:
            if isinstance(logger, LoguruLogger):
                return StateOnLoguru
        return StateOnUser

    def info(self, msg: t.Any) -> None:
        self.__state.info(self, msg)

    def error(self, msg: t.Any) -> None:
        self.__state.error(self, msg)

    def debug_request(self, request: object) -> None:
        self.__state.debug_request(self, request)

    def debug_response(self, response: object) -> None:
        self.__state.debug_response(self, response)

    def debug_push(self, response: object) -> None:
        self.__state.debug_push(self, response)


class CustomLogger(t.Protocol):
    """
        The custom logger must be an instance of python Logger or Loguru Logger.
        by providing the custom class of the logger we will not waranty as all of 
        the messages will be visible to the user.
    """
    def debug(self, msg, /, *args, **kwargs: t.Any) -> None:
        ...

    def info(self, msg, /, *args, **kwargs: t.Any) -> None:
        ...

    def warning(self, msg, /, *args, **kwargs: t.Any) -> None:
        ...

    def error(self, msg, /, *args, **kwargs: t.Any) -> None:
        ...


class LoggerState(t.Protocol):
    @staticmethod
    def setup(connid: str, logger) -> CustomLogger | None:
        return None

    @staticmethod
    def info(inst: "TransportationLogger", msg: t.Any) -> None:
        ...

    @staticmethod
    def error(inst: "TransportationLogger", msg: t.Any) -> None:
        ...

    @staticmethod
    def debug_request(inst: "TransportationLogger", request: object) -> None:
        ...

    @staticmethod
    def debug_response(inst: "TransportationLogger", response: object) -> None:
        ...

    @staticmethod
    def debug_push(inst: "TransportationLogger", response: object) -> None:
        ...



from __future__ import annotations
import sys
from functools import partial
from typing import (
    Dict,
    ForwardRef,
    Iterable,
    Type,
    Any,
    Optional,
    _eval_type,  # type: ignore
)


def resolve_annotations(raw_annotations: dict[str, Type[Any]], module_name: str | None) -> dict[str, Type[Any]]:
    """
    Partially taken from typing.get_type_hints.

    Resolve string or ForwardRef annotations into type objects if possible.
    """
    base_globals: Optional[Dict[str, Any]] = None
    if module_name:
        if module := sys.modules.get(module_name, None):
            base_globals = module.__dict__

    validate_version = (3, 10) > sys.version_info >= (3, 9, 8) or sys.version_info >= (3, 10, 1)
    forward_ref = (
        partial(ForwardRef, is_argument=False, is_class=True)
        if validate_version else
        partial(ForwardRef, is_argument=False)
    )

    annotations = {}
    for name, value in raw_annotations.items():
        val_ = forward_ref(value) if isinstance(value, str) else value
        try:
            val_ = _eval_type(val_, base_globals, None)
        except NameError:
            pass
        except TypeError:
            if sys.version_info < (3, 9):
                version_str = f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
                raise TypeError(f"The Python version you using is: {version_str} and it's supports type hints only from <typing> module") from None
        annotations[name] = val_
    return annotations


def flatten(items, ignore_types=(str, bytes)):
    for x in items:
        if isinstance(x, Iterable) and not isinstance(x, ignore_types):
            yield from flatten(x)
        else:
            yield x



from typing import (
    TYPE_CHECKING,
    Tuple,
)
if TYPE_CHECKING:
    from xoa_driver.internals.core import interfaces as itf
from xoa_driver.internals.commands import (
    PP_PRECODING,
    PP_PRECODINGSTATUS,
    PP_GRAYCODING,
    PL1_AUTONEGINFO,
    PL1_LINKTRAININFO,
    PL1_LOG,
    PL1_CFG_TMP,
    PL1_LINKTRAIN_CMD,
    PL1_CTRL,
    PL1_GET_DATA,
    PL1_PHYTXEQ_LEVEL,
    PL1_PHYTXEQ_COEFF,
    PL1_AUTONEG_STATUS,
    PL1_AUTONEG_ABILITIES,
    PL1_PCS_VARIANT,
    PL1_AUTONEG_CONFIG,
    PL1_ANLT,
    PL1_PHYTXEQ,
    PL1_LINKTRAIN_CONFIG,
    PL1_LINKTRAIN_STATUS,
    PP_PHYRXEQ_EXT,
    PP_PHYRXEQSTATUS_EXT
)
from .pcs_pma_ghijkl import (
    Prbs,
)
from xoa_driver import enums

class FreyaPMA:
    """Freya PMA"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, serdes_xindex: int) -> None:
        self.precoding = PP_PRECODING(conn, module_id, port_id, serdes_xindex)
        """GET/SET Pre-Coding Configurations. (only for Freya)

        :type: PP_PRECODING
        """

        self.graycoding = PP_GRAYCODING(conn, module_id, port_id, serdes_xindex)
        """GET/SET Gray-Coding Configurations. (only for Freya)

        :type: PP_GRAYCODING
        """

class FreyaSIV:
    """Freya Signal Integrity View"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, serdes_xindex: int) -> None:
        self.control = PL1_CTRL(conn, module_id, port_id, serdes_xindex, enums.Layer1Control.SAMPLED_SIGNAL_INTEGRITY_SCAN)
        """Control SIV scan. (only for Freya)

        :type: PL1_CTRL
        """

        self.data = PL1_GET_DATA(conn, module_id, port_id, serdes_xindex, enums.Layer1Control.SAMPLED_SIGNAL_INTEGRITY_SCAN)
        """Get SIV scan data. (only for Freya)

        :type: PL1_GET_DATA
        """

class FreyaTxTap:
    """Freya Tx Tap"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, serdes_xindex: int) -> None:
        self.native = PL1_PHYTXEQ(conn, module_id, port_id, serdes_xindex)
        """TX tap native value. (only for Freya)

        :type: PL1_PHYTXEQ
        """

        self.level = PL1_PHYTXEQ_LEVEL(conn, module_id, port_id, serdes_xindex)
        """TX tap mV/dB value. (only for Freya)

        :type: PL1_PHYTXEQ_LEVEL
        """

        self.ieee = PL1_PHYTXEQ_COEFF(conn, module_id, port_id, serdes_xindex)
        """TX tap IEEE coefficient value. (only for Freya)

        :type: PL1_PHYTXEQ_COEFF
        """

class FreyaRxTapConfig:
    """Freya Rx Tap Configuration"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, serdes_xindex: int) -> None:
        self.ctle_low = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.CTLE_LOW)
        """RX tap CTLE LOW. (only for Freya)
        """

        self.ctle_high = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.CTLE_HIGH)
        """RX tap CTLE HIGH. (only for Freya)
        """

        self.agc = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.AGC)
        """RX tap Automatic Gain Control. (only for Freya)
        """

        self.oc = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.OC)
        """RX tap Offset Cancellation. (only for Freya)
        """

        self.cdr = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.CDR)
        """RX tap Clock and Data Recovery. (only for Freya)
        """

        self.pre_ffe_1 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.PRE_FFE_1)
        """RX tap Pre Feed-Forward Equalizer #1. (only for Freya)
        """

        self.pre_ffe_2 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.PRE_FFE_2)
        """RX tap Pre Feed-Forward Equalizer #2. (only for Freya)
        """

        self.pre_ffe_3 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.PRE_FFE_3)
        """RX tap Pre Feed-Forward Equalizer #3. (only for Freya)
        """

        self.pre_ffe_4 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.PRE_FFE_4)
        """RX tap Pre Feed-Forward Equalizer #4. (only for Freya)
        """

        self.pre_ffe_5 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.PRE_FFE_5)
        """RX tap Pre Feed-Forward Equalizer #5. (only for Freya)
        """

        self.pre_ffe_6 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.PRE_FFE_6)
        """RX tap Pre Feed-Forward Equalizer #6. (only for Freya)
        """

        self.pre_ffe_7 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.PRE_FFE_7)
        """RX tap Pre Feed-Forward Equalizer #7. (only for Freya)
        """

        self.pre_ffe_8 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.PRE_FFE_8)
        """RX tap Pre Feed-Forward Equalizer #8. (only for Freya)
        """

        self.dfe = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.DFE)
        """RX tap Decision Feedback Equalization. (only for Freya)
        """

        self.post_ffe_1 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_1)
        """RX tap Post Feed-Forward Equalizer #1. (only for Freya)
        """

        self.post_ffe_2 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_2)
        """RX tap Post Feed-Forward Equalizer #2. (only for Freya)
        """

        self.post_ffe_3 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_3)
        """RX tap Post Feed-Forward Equalizer #3. (only for Freya)
        """

        self.post_ffe_4 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_4)
        """RX tap Post Feed-Forward Equalizer #4. (only for Freya)
        """

        self.post_ffe_5 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_5)
        """RX tap Post Feed-Forward Equalizer #5. (only for Freya)
        """

        self.post_ffe_6 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_6)
        """RX tap Post Feed-Forward Equalizer #6. (only for Freya)
        """

        self.post_ffe_7 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_7)
        """RX tap Post Feed-Forward Equalizer #7. (only for Freya)
        """

        self.post_ffe_8 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_8)
        """RX tap Post Feed-Forward Equalizer #8. (only for Freya)
        """

        self.post_ffe_9 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_9)
        """RX tap Post Feed-Forward Equalizer #9. (only for Freya)
        """

        self.post_ffe_10 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_10)
        """RX tap Post Feed-Forward Equalizer #10. (only for Freya)
        """

        self.post_ffe_11 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_11)
        """RX tap Post Feed-Forward Equalizer #11. (only for Freya)
        """

        self.post_ffe_12 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_12)
        """RX tap Post Feed-Forward Equalizer #12. (only for Freya)
        """

        self.post_ffe_13 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_13)
        """RX tap Post Feed-Forward Equalizer #13. (only for Freya)
        """

        self.post_ffe_14 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_14)
        """RX tap Post Feed-Forward Equalizer #14. (only for Freya)
        """

        self.post_ffe_15 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_15)
        """RX tap Post Feed-Forward Equalizer #15. (only for Freya)
        """

        self.post_ffe_16 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_16)
        """RX tap Post Feed-Forward Equalizer #16. (only for Freya)
        """

        self.post_ffe_17 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_17)
        """RX tap Post Feed-Forward Equalizer #17. (only for Freya)
        """

        self.post_ffe_18 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_18)
        """RX tap Post Feed-Forward Equalizer #18. (only for Freya)
        """

        self.post_ffe_19 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_19)
        """RX tap Post Feed-Forward Equalizer #19. (only for Freya)
        """

        self.post_ffe_20 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_20)
        """RX tap Post Feed-Forward Equalizer #20. (only for Freya)
        """

        self.post_ffe_21 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_21)
        """RX tap Post Feed-Forward Equalizer #21. (only for Freya)
        """

        self.post_ffe_22 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_22)
        """RX tap Post Feed-Forward Equalizer #22. (only for Freya)
        """

        self.post_ffe_23 = PP_PHYRXEQ_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_13)
        """RX tap Post Feed-Forward Equalizer #23. (only for Freya)
        """

class FreyaRxTapStatus:
    """Freya Rx Tap Status"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, serdes_xindex: int) -> None:
        self.ctle_low = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.CTLE_LOW)
        """RX tap CTLE LOW. (only for Freya)
        """

        self.ctle_high = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.CTLE_HIGH)
        """RX tap CTLE HIGH. (only for Freya)
        """

        self.agc = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.AGC)
        """RX tap Automatic Gain Control. (only for Freya)
        """

        self.oc = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.OC)
        """RX tap Offset Cancellation. (only for Freya)
        """

        self.cdr = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.CDR)
        """RX tap Clock and Data Recovery. (only for Freya)
        """

        self.pre_ffe_1 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.PRE_FFE_1)
        """RX tap Pre Feed-Forward Equalizer #1. (only for Freya)
        """

        self.pre_ffe_2 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.PRE_FFE_2)
        """RX tap Pre Feed-Forward Equalizer #2. (only for Freya)
        """

        self.pre_ffe_3 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.PRE_FFE_3)
        """RX tap Pre Feed-Forward Equalizer #3. (only for Freya)
        """

        self.pre_ffe_4 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.PRE_FFE_4)
        """RX tap Pre Feed-Forward Equalizer #4. (only for Freya)
        """

        self.pre_ffe_5 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.PRE_FFE_5)
        """RX tap Pre Feed-Forward Equalizer #5. (only for Freya)
        """

        self.pre_ffe_6 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.PRE_FFE_6)
        """RX tap Pre Feed-Forward Equalizer #6. (only for Freya)
        """

        self.pre_ffe_7 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.PRE_FFE_7)
        """RX tap Pre Feed-Forward Equalizer #7. (only for Freya)
        """

        self.pre_ffe_8 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.PRE_FFE_8)
        """RX tap Pre Feed-Forward Equalizer #8. (only for Freya)
        """

        self.dfe = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.DFE)
        """RX tap Decision Feedback Equalization. (only for Freya)
        """

        self.post_ffe_1 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_1)
        """RX tap Post Feed-Forward Equalizer #1. (only for Freya)
        """

        self.post_ffe_2 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_2)
        """RX tap Post Feed-Forward Equalizer #2. (only for Freya)
        """

        self.post_ffe_3 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_3)
        """RX tap Post Feed-Forward Equalizer #3. (only for Freya)
        """

        self.post_ffe_4 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_4)
        """RX tap Post Feed-Forward Equalizer #4. (only for Freya)
        """

        self.post_ffe_5 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_5)
        """RX tap Post Feed-Forward Equalizer #5. (only for Freya)
        """

        self.post_ffe_6 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_6)
        """RX tap Post Feed-Forward Equalizer #6. (only for Freya)
        """

        self.post_ffe_7 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_7)
        """RX tap Post Feed-Forward Equalizer #7. (only for Freya)
        """

        self.post_ffe_8 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_8)
        """RX tap Post Feed-Forward Equalizer #8. (only for Freya)
        """

        self.post_ffe_9 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_9)
        """RX tap Post Feed-Forward Equalizer #9. (only for Freya)
        """

        self.post_ffe_10 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_10)
        """RX tap Post Feed-Forward Equalizer #10. (only for Freya)
        """

        self.post_ffe_11 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_11)
        """RX tap Post Feed-Forward Equalizer #11. (only for Freya)
        """

        self.post_ffe_12 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_12)
        """RX tap Post Feed-Forward Equalizer #12. (only for Freya)
        """

        self.post_ffe_13 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_13)
        """RX tap Post Feed-Forward Equalizer #13. (only for Freya)
        """

        self.post_ffe_14 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_14)
        """RX tap Post Feed-Forward Equalizer #14. (only for Freya)
        """

        self.post_ffe_15 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_15)
        """RX tap Post Feed-Forward Equalizer #15. (only for Freya)
        """

        self.post_ffe_16 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_16)
        """RX tap Post Feed-Forward Equalizer #16. (only for Freya)
        """

        self.post_ffe_17 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_17)
        """RX tap Post Feed-Forward Equalizer #17. (only for Freya)
        """

        self.post_ffe_18 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_18)
        """RX tap Post Feed-Forward Equalizer #18. (only for Freya)
        """

        self.post_ffe_19 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_19)
        """RX tap Post Feed-Forward Equalizer #19. (only for Freya)
        """

        self.post_ffe_20 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_20)
        """RX tap Post Feed-Forward Equalizer #20. (only for Freya)
        """

        self.post_ffe_21 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_21)
        """RX tap Post Feed-Forward Equalizer #21. (only for Freya)
        """

        self.post_ffe_22 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_22)
        """RX tap Post Feed-Forward Equalizer #22. (only for Freya)
        """

        self.post_ffe_23 = PP_PHYRXEQSTATUS_EXT(conn, module_id, port_id, serdes_xindex, enums.RxEqExtCap.POST_FFE_13)
        """RX tap Post Feed-Forward Equalizer #23. (only for Freya)
        """

class FreyaRxTap:
    """Freya Rx tap
    """
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, serdes_xindex: int) -> None:
        self.config = FreyaRxTapConfig(conn, module_id, port_id, serdes_xindex)
        """Freya Rx tap configuration
        """
        self.status = FreyaRxTapStatus(conn, module_id, port_id, serdes_xindex)
        """Freya Rx tap status
        """

class FreyaMedium:
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, serdes_xindex: int) -> None:
        self.tx = FreyaTxTap(conn, module_id, port_id, serdes_xindex)
        """Freya Tx tap
        """
        self.rx = FreyaRxTap(conn, module_id, port_id, serdes_xindex)
        """Freya Rx tap
        """
        self.siv = FreyaSIV(conn, module_id, port_id, serdes_xindex)
        """Freya Signal Integrity
        """

class SerDesFreya:
    """L23 high-speed port SerDes configuration and status."""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int, serdes_xindex: int) -> None:

        self.prbs = Prbs(conn, module_id, port_id, serdes_xindex)
        """PRBS
        :type: Prbs
        """
        
        self.pma = FreyaPMA(conn, module_id, port_id, serdes_xindex)
        """Freya PMA

        :type: FreyaPMA
        """

        self.medium = FreyaMedium(conn, module_id, port_id, serdes_xindex)
        """Freya medium

        :type: FreyaMedium
        """

        self.lt_cmd = PL1_LINKTRAIN_CMD(conn, module_id, port_id, serdes_xindex)
        """Link training command.
        
        :type: PP_LINKTRAIN
        """

        self.lt_info = PL1_LINKTRAININFO(conn, module_id, port_id, serdes_xindex, 0)
        """Link training info.
        
        :type: PL1_LINKTRAININFO
        """

        self.lt_status = PL1_LINKTRAIN_STATUS(conn, module_id, port_id, serdes_xindex)
        """Link training status.
        
        :type: PL1_LINKTRAIN_STATUS
        """

class FreyaAutoNeg:
    """Freya Autoneg"""

    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.info = PL1_AUTONEGINFO(conn, module_id, port_id, 0)
        """Autoneg info

        :type: PL1_AUTONEGINFO
        """

        self.status = PL1_AUTONEG_STATUS(conn, module_id, port_id)
        """Autoneg status

        :type: PL1_AUTONEG_STATUS
        """

        self.abilities = PL1_AUTONEG_ABILITIES(conn, module_id, port_id)
        """Autoneg abilities

        :type: PL1_AUTONEG_ABILITIES
        """

        self.config = PL1_AUTONEG_CONFIG(conn, module_id, port_id)
        """Autoneg configuration

        :type: PL1_AUTONEG_CONFIG
        """

class FreyaANLT:
    """Freya port-level anlt. For per-serdes configuration and status, use serdes[x]
    """
    def __init__(self, conn: "itf.IConnection", module_id: int, port_id: int) -> None:
        self.an = FreyaAutoNeg(conn, module_id, port_id)
        """Autoneg config and status
        """
        self.ctrl = PL1_ANLT(conn, module_id, port_id)
        """ANLT control
        """
        self.lt_config = PL1_LINKTRAIN_CONFIG(conn, module_id, port_id)
        """Port-level Link Training config
        """
        self.log = PL1_LOG(conn, module_id, port_id)
        """ANLT log
        """

class Layer1:
    def __init__(self, conn: "itf.IConnection", port) -> None:
        self.serdes: Tuple[SerDesFreya, ...] = tuple(
                SerDesFreya(conn, *port.kind, serdes_xindex=idx)
                for idx in range(port.info.capabilities.serdes_count)
                )
        self.anlt = FreyaANLT(conn, *port.kind)
        """Freya port-level anlt. For per-serdes configuration and status, use serdes[x]
        """
        self.pcs_variant = PL1_PCS_VARIANT(conn, *port.kind)
        """PCS variant configuration
        """
        




#: All exception classes which can be propagated to the upper level.

from .internals.exceptions import (
    WrongModuleError,
    WrongTesterError,
    WrongTesterPasswordError,
)
from .internals.core.exceptions import (
    RepeatedRequestID,
    TransporterException,
    XmpBadCommandError,
    XmpBadHeaderError,
    XmpBadIndexError,
    XmpBadModuleError,
    XmpBadParameterError,
    XmpBadPortError,
    XmpBadSizeError,
    XmpBadValueError,
    XmpFailedError,
    XmpMemoryFailureError,
    XmpModuleOperationNotSupportedByChassisError,
    XmpNoConnectionError,
    XmpNoFreePortLicenseError,
    XmpNoLoggedOnError,
    XmpNoPeLicenseError,
    XmpNotReadableError,
    XmpNotReservedError,
    XmpNotSupportedError,
    XmpNotValidError,
    XmpNotWritableError,
    XmpPendingError,
    XmpStatusException,
    XmpXlsDeniedError,
    XmpXlsFailedError,
    XmpXlsInvalidError,
    XoaConnectionError,
    XoaConnectionTimeoutError,
    XoaException,
    XoaLostFuture,
)

__all__ = (
    "RepeatedRequestID",
    "TransporterException",
    "WrongModuleError",
    "WrongTesterError",
    "WrongTesterPasswordError",
    "XmpBadCommandError",
    "XmpBadHeaderError",
    "XmpBadIndexError",
    "XmpBadModuleError",
    "XmpBadParameterError",
    "XmpBadPortError",
    "XmpBadSizeError",
    "XmpBadValueError",
    "XmpFailedError",
    "XmpMemoryFailureError",
    "XmpModuleOperationNotSupportedByChassisError",
    "XmpNoConnectionError",
    "XmpNoFreePortLicenseError",
    "XmpNoLoggedOnError",
    "XmpNoPeLicenseError",
    "XmpNotReadableError",
    "XmpNotReservedError",
    "XmpNotSupportedError",
    "XmpNotValidError",
    "XmpNotWritableError",
    "XmpPendingError",
    "XmpStatusException",
    "XmpXlsDeniedError",
    "XmpXlsFailedError",
    "XmpXlsInvalidError",
    "XoaConnectionError",
    "XoaConnectionTimeoutError",
    "XoaException",
    "XoaLostFuture",
)



from __future__ import annotations
from dataclasses import dataclass
import typing
import functools

from xoa_driver.internals.core.builders import (
    build_get_request,
    build_set_request
)
from xoa_driver.internals.core import interfaces
from xoa_driver.internals.core.token import Token
from xoa_driver.internals.core.transporter.registry import register_command
from xoa_driver.internals.core.transporter.protocol.payload import (
    field,
    RequestBodyStruct,
    ResponseBodyStruct,
    XmpByte,
    XmpInt,
    XmpLong,
    XmpSequence,

)
from .enums import (
    OnOff,
    SourceType,
    PacketDetailSelection,
)


@register_command
@dataclass
class PD_INDICES:
    """
    Obtain or configure histogram indices for each of N histograms.
    """

    code: typing.ClassVar[int] = 140
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int

    class GetDataAttr(ResponseBodyStruct):
        histogram_indices: typing.List[int] = field(XmpSequence(types_chunk=[XmpInt()]))
        """list of integers, histogram indices"""

    class SetDataAttr(RequestBodyStruct):
        histogram_indices: typing.List[int] = field(XmpSequence(types_chunk=[XmpInt()]))
        """list of integers, histogram indices"""

    def get(self) -> Token[GetDataAttr]:
        """Get the histogram indices for each of N histograms.

        :return: the histogram indices for each of N histograms
        :rtype: PD_INDICES.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port))

    def set(self, histogram_indices: typing.List[int]) -> Token[None]:
        """Set the histogram indices for each of N histograms.

        :param histogram_indices: histogram indices
        :type histogram_indices: typing.List[int]
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, histogram_indices=histogram_indices))


@register_command
@dataclass
class PD_CREATE:
    """
    Creates a histogram definition with the specified sub-index value.
    """

    code: typing.ClassVar[int] = 141
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _dataset_xindex: int

    class SetDataAttr(RequestBodyStruct):
        pass

    def set(self) -> Token[None]:
        """Create a histogram definition with the specified sub-index value.
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._dataset_xindex]))


@register_command
@dataclass
class PD_DELETE:
    """
    Delete an existing histogram definition.
    """

    code: typing.ClassVar[int] = 142
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _dataset_xindex: int

    class SetDataAttr(RequestBodyStruct):
        pass

    def set(self) -> Token[None]:
        """Delete an existing histogram definition.
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._dataset_xindex]))


@register_command
@dataclass
class PD_ENABLE:
    """
    Whether a histogram is currently active on a port. When turned on, all the bucket
    counts are cleared to zero. Subsequently each packet matching the histogram source
    criteria is counted into one of the buckets. While a histogram is enabled its
    parameters cannot be changed.
    """

    code: typing.ClassVar[int] = 143
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _dataset_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        on_off: OnOff = field(XmpByte())
        """coded integer, whether the histogram is enabled."""

    class SetDataAttr(RequestBodyStruct):
        on_off: OnOff = field(XmpByte())
        """coded integer, whether the histogram is enabled."""

    def get(self) -> Token[GetDataAttr]:
        """Get whether a histogram is currently active on a port.

        :return: whether the histogram is enabled
        :rtype: OnOff
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._dataset_xindex]))

    def set(self, on_off: OnOff) -> Token[None]:
        """Set whether a histogram is currently active on a port.

        :param on_off: whether the histogram is enabled
        :type on_off: OnOff
        """

        return Token(self._connection, build_set_request(self, module=self._module, port=self._port, indices=[self._dataset_xindex], on_off=on_off))

    set_off = functools.partialmethod(set, OnOff.OFF)
    """Disable the histogram on the port.
    """

    set_on = functools.partialmethod(set, OnOff.ON)
    """Enable the histogram on the port.
    """


@register_command
@dataclass
class PD_SOURCE:
    """
    The source criteria specifying what is counted, and for which packets, by a
    histogram of a port.
    """

    code: typing.ClassVar[int] = 144
    pushed: typing.ClassVar[bool] = True

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _dataset_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        source_type: SourceType = field(XmpInt())
        """coded integer, specifying what is counted and for which packets."""
        which_packets: PacketDetailSelection = field(XmpInt())
        """coded integer, specifying a further detail on which packets to count."""
        identity: int = field(XmpInt())
        """integer, test payload id or filter id for the wanted packets."""

    class SetDataAttr(RequestBodyStruct):
        source_type: SourceType = field(XmpInt())
        """coded integer, specifying what is counted and for which packets."""
        which_packets: PacketDetailSelection = field(XmpInt())
        """coded integer, specifying a further detail on which packets to count."""
        identity: int = field(XmpInt())
        """integer, test payload id or filter id for the wanted packets."""

    def get(self) -> Token[GetDataAttr]:
        """Get the source criteria specifying what is counted, and for which packets, by a histogram of a port.

        :return: what is counted and for which packets, a further detail on which packets to count, test payload id or filter id for the wanted packets
        :rtype: PD_SOURCE.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._dataset_xindex]))

    def set(self, source_type: SourceType, which_packets: PacketDetailSelection, identity: int) -> Token[None]:
        """Set the source criteria specifying what is counted, and for which packets, by a histogram of a port.

        :param source_type: what is counted and for which packets
        :type source_type: SourceType
        :param which_packets: a further detail on which packets to count
        :type which_packets: PacketDetailSelection
        :param identity: test payload id or filter id for the wanted packets
        :type identity: int
        """

        return Token(
            self._connection,
            build_set_request(
                self,
                module=self._module,
                port=self._port,
                indices=[self._dataset_xindex],
                source_type=source_type,
                which_packets=which_packets,
                identity=identity
            )
        )


@register_command
@dataclass
class PD_RANGE:
    """
    The bucket ranges used for classifying the packets counted by a histogram of a
    port. The packets are either counted by length, measured in bytes, by inter-
    frame gap to the preceding packet,  also measured in bytes, or by latency in
    transmission measured in nanoseconds.  There are a fixed number of buckets, each
    middle bucket covering a fixed-size range of values which is a power of two.
    The first and last buckets count all the packets that do not fit within the
    ranges of the middle buckets. The buckets are placed at a certain offset by
    specifying the first value that should be counted by the first middle bucket.
    """

    code: typing.ClassVar[int] = 145
    pushed: typing.ClassVar[bool] = True

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _dataset_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        start: int = field(XmpInt())
        """integer, first value going into the second bucket."""
        step: int = field(XmpInt())
        """integer, the span of each middle bucket:
            (1) 1,2,4,8,16,32,64,128,256,512 (bytes, non-latency histograms).
            (2) 16,32,64,128,...,1048576,2097152 (nanoseconds, latency histograms).
        """
        bucket_count: int = field(XmpInt())
        """integer, the total number of buckets."""

    class SetDataAttr(RequestBodyStruct):
        start: int = field(XmpInt())
        """integer, first value going into the second bucket."""
        step: int = field(XmpInt())
        """integer, the span of each middle bucket:
            (1) 1,2,4,8,16,32,64,128,256,512 (bytes, non-latency histograms).
            (2) 16,32,64,128,...,1048576,2097152 (nanoseconds, latency histograms).
        """
        bucket_count: int = field(XmpInt())
        """integer, the total number of buckets."""

    def get(self) -> Token[GetDataAttr]:
        """Get the bucket ranges used for classifying the packets counted by a histogram of a port.

        :return: first value going into the second bucket, the span of each middle bucket, the total number of buckets
        :rtype: PD_RANGE.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._dataset_xindex]))

    def set(self, start: int, step: int, bucket_count: int) -> Token[None]:
        """Set the bucket ranges used for classifying the packets counted by a histogram of a port.
        The packets are either counted by length, measured in bytes, by inter-
        frame gap to the preceding packet,  also measured in bytes, or by latency in
        transmission measured in nanoseconds.  There are a fixed number of buckets, each
        middle bucket covering a fixed-size range of values which is a power of two.
        The first and last buckets count all the packets that do not fit within the
        ranges of the middle buckets. The buckets are placed at a certain offset by
        specifying the first value that should be counted by the first middle bucket.

        :param start: first value going into the second bucket
        :type start: int
        :param step: the span of each middle bucket:
            (1) 1,2,4,8,16,32,64,128,256,512 (bytes, non-latency histograms).
            (2) 16,32,64,128,...,1048576,2097152 (nanoseconds, latency histograms)
        :type step: int
        :param bucket_count: the total number of buckets
        :type bucket_count: int
        """

        return Token(
            self._connection,
            build_set_request(
                self,
                module=self._module,
                port=self._port,
                indices=[self._dataset_xindex],
                start=start,
                step=step,
                bucket_count=bucket_count
            )
        )


@register_command
@dataclass
class PD_SAMPLES:
    """
    The current set of counts collected by a histogram for a port. There is one value
    for each bucket, but any trailing zeros are left out. The list is empty if all
    counts are zero.
    """

    code: typing.ClassVar[int] = 146
    pushed: typing.ClassVar[bool] = False

    _connection: 'interfaces.IConnection'
    _module: int
    _port: int
    _dataset_xindex: int

    class GetDataAttr(ResponseBodyStruct):
        packet_counts: typing.List[int] = field(XmpSequence(types_chunk=[XmpLong()]))
        """list of long integers, the number of packets counted for each bucket."""

    def get(self) -> Token[GetDataAttr]:
        """Get current set of counts collected by a histogram for a port.

        :return: the number of packets counted for each bucket
        :rtype: PD_SAMPLES.GetDataAttr
        """

        return Token(self._connection, build_get_request(self, module=self._module, port=self._port, indices=[self._dataset_xindex]))



#: Low-Level interface


from xoa_driver.internals import commands
from xoa_driver.internals.core.transporter.registry import get_command
from xoa_driver.internals.core.transporter.handler import TransportationHandler
from xoa_driver.internals.core.funcs import establish_connection


__all__ = (
    "commands",
    "get_command",
    "TransportationHandler",
    "establish_connection"
)
